# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test LDAP Text attribute type """

import pytest

from eesyldap.attributes import Text

# pylint: disable=protected-access,redefined-outer-name


@pytest.fixture
def fake_attr():
    """ Generate a fake LDAP Text attribute """
    return Text('fakeAttr')


def test_to_python(fake_attr):
    """ Test LDAP Text attribute _to_python method """
    assert fake_attr._to_python(b'ascii') == 'ascii'
    assert fake_attr._to_python('ÀbĈ'.encode('utf8')) == u'ÀbĈ'
    assert fake_attr._to_python(None) is None


def test_to_python_sortable(fake_attr):
    """ Test LDAP Text attribute _to_python method with sortable=True """
    assert fake_attr._to_python(None, sortable=True) == ''
    assert fake_attr._to_python('ascii'.encode('utf8'), sortable=True) == 'ascii'


def test_to_ldap(fake_attr):
    """ Test LDAP Text attribute _to_ldap method """
    assert fake_attr._to_ldap('ascii') == b'ascii'
    assert fake_attr._to_ldap('ÀbĈ') == 'ÀbĈ'.encode('utf8')
    with pytest.raises(AssertionError):
        fake_attr._to_ldap(None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Helpers """

import ldap


def format_modlist(modlist, prefix=None):
    """
    Format modlist as generated by ldap.modlist functions

    :param  modlist:    The modlist to format
    :param  prefix:     Prefix string to add on each output lines (optional)
    """
    assert isinstance(modlist, list), "modlist must be a list as generated by ldap.modlist functions (not %s)" % type(modlist)
    if not modlist:
        return "No change"
    msg = []
    if not prefix:
        prefix = ''
    assert isinstance(prefix, str), "prefix must be a string (not %s)" % type(prefix)
    for (op, attr, val) in modlist:
        if op == ldap.MOD_ADD:  # pylint: disable=no-member
            op = 'ADD'
        elif op == ldap.MOD_DELETE:  # pylint: disable=no-member
            op = 'DELETE'
        elif op == ldap.MOD_REPLACE:  # pylint: disable=no-member
            op = 'REPLACE'
        else:
            op = 'UNKNOWN (=%s)' % op
        if val is None and op == 'DELETE':
            msg.append('%s - %s %s' % (prefix, op, attr))
        else:
            msg.append('%s - %s %s: %s' % (prefix, op, attr, val))
    return '\n'.join(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Object """

import copy
import inspect
import logging

from eesyldap.entry import LDAPEntry
from eesyldap.exceptions import LDAPDuplicatedRelationName
from eesyldap.exceptions import LDAPInvalidAttributeName
from eesyldap.exceptions import LDAPRelationAttributeConflictName
from eesyldap.filters import combine as combine_ldap_filter

log = logging.getLogger(__name__)


class LDAPObject:
    """
    LDAP object

    This class implement a python object behind a LDAP entry
    """

    # List of the LDAP objectClass
    _class = []

    # Dict of the python object attributes
    _attrs = {}

    # Reversed mapping dict of the LDAP attribute to the python object attribute
    _srtta = None

    # RDN python and LDAP attribute
    _rdn_attr = None
    _rdn_ldap_attr = None

    # Protected attributes
    # List of attributes that will never be modified.
    _protected_attrs = None

    # Relations
    _relations = {}

    # LDAPClient object reference
    _client = None

    # LDAPEntry object reference
    _entry = None

    # Iterator index
    __iter_idx = 0

    def __init__(self, client, dn=None, attrs_values=None, entry=None, serialized_state=None):
        """
        LDAP obect constructor

        :param  client:             The LDAPClient object
        :param  dn:                 The LDAP object DN (optional for new object)
        :param  attrs_values:       The LDAP object attributes's values (optional for new object)
        :param  entry:              The LDAP entry object (optional, could be specified instead of
                                    dn and attrs_values parameters)
        :param  serialized_state:   The LDAP object serialized state (optional, could be specified
                                    instead of dn and attrs_values parameters)
        """
        self._client = client
        log.debug('Instanciate new %s object with dn %s', self.__class__, dn)
        if not self._class:
            raise NotImplementedError("%s : You must defined _class attribute" % self.__class__.__name__)
        if not self._attrs:
            raise NotImplementedError("%s : You must defined _attrs attribute" % self.__class__.__name__)
        if not isinstance(self._rdn_ldap_attr, str) and not isinstance(self._rdn_attr, str):
            raise NotImplementedError("%s : You must defined _rdn_attr or _rdn_ldap_attr attribute" % self.__class__.__name__)

        # Generate reversed mapping dict of LDAP attribute to Python object attribute
        self._srtta = self._get_srtta()

        # Check RDN attribute
        self._rdn_ldap_attr = self.get_rdn_ldap_attr()
        self._rdn_attr = self.get_rdn_attr()
        if self.is_multiple(self._rdn_attr):
            raise NotImplementedError('%s : The attribute %s defined as RDN is multiple and multi-valued attribute is not yet supported as RDN' % (self.__class__.__name__, self._rdn_attr))

        # Check relations
        for rel_name in self._relations:
            # Check conflict between attribute and relation names
            if rel_name in self._attrs:
                raise LDAPRelationAttributeConflictName(self.__class__, rel_name, self._attrs[rel_name])

        # Create dedicated LDAP entry
        if isinstance(entry, LDAPEntry):
            self._entry = entry
        elif serialized_state is not None:
            self._entry = LDAPEntry.unserialize(client, self, serialized_state)
        else:
            self._entry = LDAPEntry(client, self, dn, attrs_values)

        # Iterator index
        self.__iter_idx = 0

    @classmethod
    def _get_srtta(cls):
        """ Generate reversed mapping dict of LDAP attribute to Python object attribute """
        srtta = {}
        for attr in cls._attrs:
            ldap_attr_name = cls._attrs[attr].ldap_name
            assert ldap_attr_name not in srtta, "LDAP attribute %s is affected to multiple object attributes : %s / %s" % (ldap_attr_name, srtta[ldap_attr_name], attr)
            srtta[ldap_attr_name] = attr
        return srtta

    def get_dn(self):
        """ Retreive or generate LDAP object DN """
        return self._entry.get_dn()

    def __getattr__(self, name):
        """ Get an DN or attributes's value """
        if name == 'dn':
            return self._entry.get_dn()
        if name == 'attrs':
            return LDAPObjectAttributesWrapper(self, self._attrs)
        if name == 'classes':
            return self._entry.get_values('objectClass', default=self._class)
        if name == 'entry':
            return self._entry
        if name == 'client':
            return self._client
        if name == 'protected_attrs':
            return copy.copy(self._protected_attrs) or []
        if name == 'rdn_attr':
            return copy.copy(self._rdn_attr)
        if name == 'rdn_ldap_attr':
            return copy.copy(self._rdn_ldap_attr)
        if self.exists(name):
            return self._attrs[name].get_value(self)
        if name in self._relations:
            return self._relations[name].get_related_objects(self)
        raise LDAPInvalidAttributeName(name)

    def __setattr__(self, name, value):
        """ Set an attribute value """
        if name in self._attrs:
            return self._attrs[name].set_value(self, value)
        if name in self._relations:
            return self._relations[name].set_related_objects(self, value)
        if hasattr(self, name):
            return object.__setattr__(self, name, value)
        raise LDAPInvalidAttributeName(name)

    def __iter__(self):
        return self

    def __next__(self):
        self.__iter_idx += 1
        try:
            return self._attrs.keys()[self.__iter_idx - 1]
        except IndexError:
            self.__iter_idx = 0
            raise StopIteration

    def items(self):
        """ Return LDAP object's attributes items like a dict """
        return [(attr_name, attr.get_value(self)) for attr_name, attr in self._attrs.items()]

    @classmethod
    def keys(cls):
        """ Return LDAP object's attribute names like dict().keys() """
        return cls._attrs.keys()

    @classmethod
    def ldap_attr_names(cls):
        """ Return list of the LDAP attribute names """
        return [attr.ldap_name for attr_name, attr in cls._attrs.items()]

    def __getitem__(self, key):
        return self._attrs[key].get_value(self)

    # Python 2.x compatibility
    next = __next__

    def __contains__(self, item):
        return item in self._attrs

    def __eq__(self, other):
        return self.dn == other.dn and dict(self) == dict(other)

    def is_present(self, name):
        """ Check if attribute is present/defined """
        if name in self._attrs:
            return self._attrs[name].is_present(self)
        raise LDAPInvalidAttributeName(name)

    @classmethod
    def is_unique(cls, name):
        """ Check if attribute is unique """
        if name in cls._attrs:
            return cls._attrs[name].is_unique()
        raise LDAPInvalidAttributeName(name)

    @classmethod
    def is_multiple(cls, name):
        """ Check if attribute is multiple """
        if name in cls._attrs:
            return cls._attrs[name].is_multiple()
        raise LDAPInvalidAttributeName(name)

    def is_modified(self, attr=None):
        """ Check if the object (or one of this attribute) is modified """
        return self._entry.is_modified(attr)

    @classmethod
    def exists(cls, name):
        """ Check if attribute exist """
        return name in cls._attrs

    @classmethod
    def add_relation(cls, rel_name, rel):
        """ Add a relation to the object type """
        if rel_name in cls._relations:
            if cls._relations[rel_name] == rel:
                return True
            raise LDAPDuplicatedRelationName(cls, rel_name, cls._relations[rel_name])
        if rel_name in cls._attrs:
            raise LDAPRelationAttributeConflictName(cls, rel_name, cls._attrs[rel_name])
        log.debug('Add relation %s on objects %s (%s)', rel_name, cls.__name__, rel)
        cls._relations[rel_name] = rel
        return True

    @classmethod
    def add_reverse_relations(cls):
        """ Add reverse relations """
        # Check relations
        for rel_name in cls._relations:
            cls._relations[rel_name].add_reverse_relation(cls)

    @classmethod
    def get_filter(cls, *other_filterstrs, filterstr=None, log_op=None, **filters):
        """
        Get object LDAP filter

        :param  filterstr:          A specific LDAP filter to combine with result
        :param  log_op:             The logical operator to use to combine additional filters (Default: and)
        :param  other_filterstrs:   List of filter strings
        :param  fiters:             Dict of filter attributes
        """
        filterstrs = ['(objectClass=%s)' % x for x in cls._class]
        additional_filters = []
        if filterstr:
            additional_filters.append(filterstr)
        additional_filters.extend(other_filterstrs)
        for (attr_name, attr_value) in filters.items():
            additional_filters.append(cls.get_attribute_filter(attr_name, attr_value))
        # Combine additional filters if log_op != and (default) and if we have more than one additional filters
        if len(additional_filters) > 1 and log_op is not None and log_op not in ('and', '&'):
            filterstrs.append(combine_ldap_filter(log_op, *additional_filters))
        else:
            filterstrs.extend(additional_filters)
        return combine_ldap_filter('and', *filterstrs) if len(filterstrs) > 1 else filterstrs[0]

    @classmethod
    def get_attribute_filter(cls, attr_name, attr_value=None, match=None):
        """
        Get object attribute LDAP filter string

        :param  attr_name:      The attribute name
        :param  attr_value:     The attribute value
        :param  match:          Match rule as accept by eesyldap.filters.compose (optional)
        """
        if attr_name in cls._attrs:
            return cls._attrs[attr_name].get_filter(attr_value, match=match)
        raise LDAPInvalidAttributeName(attr_name)

    @classmethod
    def get_rdn_attr(cls):
        """ Return object RDN attribute name """
        if cls._rdn_attr:
            return cls._rdn_attr
        if cls._rdn_ldap_attr:
            for attr_name, attr in cls._get_srtta().items():
                if cls._rdn_ldap_attr == attr_name:
                    cls._rdn_attr = attr
                    return attr
            raise AssertionError("Invalid LDAP attribute %s defined as RDN" % cls._rdn_ldap_attr)
        raise NotImplementedError('No RDN attribute or LDAP attribute defined')

    @classmethod
    def get_rdn_ldap_attr(cls):
        """ Return object RDN LDAP attribute name """
        if cls._rdn_ldap_attr:
            return cls._rdn_ldap_attr
        if cls._rdn_attr:
            assert cls._rdn_attr in cls._attrs, "Invalid attribute %s defined as RDN" % cls._rdn_attr
            cls._rdn_ldap_attr = cls._attrs[cls._rdn_attr].ldap_name
            return cls._rdn_ldap_attr
        raise NotImplementedError('No RDN attribute or LDAP attribute defined')

    @classmethod
    def get_python_attribute_value(cls, attr_name, ldap_attrs_values, sortable=None):
        """ Return the python value of an attribute based on specified LDAP values """
        if attr_name in cls._attrs:
            ldap_name = cls._attrs[attr_name].ldap_name.lower()
            for name, values in ldap_attrs_values.items():
                if name.lower() == ldap_name:
                    return cls._attrs[attr_name].get_value(ldap_values=values, sortable=sortable)
            # Attribute not present in LDAP values
            log.debug("Attribute %s (%s) not present in LDAP attributes's values provided (%s)", cls._attrs[attr_name], ldap_name, ldap_attrs_values)
            return cls._attrs[attr_name].get_value(ldap_values=None, sortable=sortable)
        raise LDAPInvalidAttributeName(attr_name)

    def update(self, changes=None, save=True, **to_update):
        """
        Update object attributes

        :param  changes:    Dict of changes to apply on current object
        :param  save:       Enable or disable saving changes after apply
        """
        if changes:
            assert isinstance(changes, dict), "changes parameter must be a dict"
            to_update.update(changes)

        if not to_update:
            log.debug('%s.update() : No changes', self)
            return True

        # Check changes on relations to make it only at the after saving
        changes_on_relations = {}
        for attr, value in to_update.items():
            if attr in self._relations:
                changes_on_relations[attr] = value

        if changes_on_relations and not save:
            log.error('%s : Changes on relations could only be done if save is enabled', self)
            return False

        for attr, value in to_update.items():
            if self.exists(attr):
                self.__setattr__(attr, value)
            elif attr not in changes_on_relations:
                raise LDAPInvalidAttributeName(attr)
        if save:
            if not self.save():
                return False
            # Set changes on related objects
            error = False
            for rel_name, value in changes_on_relations.items():
                if not self.__setattr__(rel_name, value):
                    error = True
            return not error
        return True

    def copy(self):
        """ Create a copy of current object """
        return self.__class__(self._client, entry=self._entry.copy())

    def save(self):
        """ Save changes on LDAP object """
        # If object have relations, keep old version of this object, before modify
        if self._relations:
            old_obj = self.copy()
            old_obj.restore()

        # Save change on LDAP entry
        if not self._entry.save():
            return False

        # If object have relation, handle changes on it relations
        if self._relations:
            error = False
            for rel_name in self._relations:
                if not self._relations[rel_name].handle_object_changes(old_obj, self):
                    error = True
            return not error
        return True

    def restore(self):
        """ Restore LDAP object at initial LDAP state """
        return self._entry.restore()

    def refresh(self):
        """ Refresh LDAP object data from LDAP """
        refreshed = self._client.get(self.__class__, dn=self.dn, refresh=True)
        if refreshed:
            self._entry = refreshed.entry
            return True
        return False

    def delete(self):
        """ Delete the LDAP object on LDAP directory """
        if not self._client.delete(self):
            return False
        if self._relations:
            error = False
            for rel_name in self._relations:
                if not self._relations[rel_name].handle_object_removal(self):
                    error = True
            return not error
        return True

    def __repr__(self):
        """ Compute and return the “official” string representation of the LDAPObject """
        return "<LDAPObject %s %s>" % (type(self).__name__, (self.dn or '[new]'))

    def serialize(self):
        """ Retreive the serialized state of the LDAP object """
        return self._entry.serialize()

    @classmethod
    def unserialize(cls, client, state):
        """ Retreived an LDAPObject from it serialized state and client object reference """
        return cls(client, serialized_state=state)


class LDAPObjectAttributesWrapper:  # pylint: disable=too-few-public-methods
    """ Wrapper on LDAP object attributes dict """

    def __init__(self, obj, attrs):
        self._obj = obj
        self._attrs = attrs

    def __getattr__(self, name):
        if name in self._attrs:
            return LDAPObjectAttributeWrapper(self._obj, self._attrs[name])
        raise LDAPInvalidAttributeName(name)


class LDAPObjectAttributeWrapper:  # pylint: disable=too-few-public-methods
    """ Wrapper on LDAP object attribute """

    def __init__(self, obj, attr):
        self.__obj = obj
        self.__attr = attr

    def __getattr__(self, name):
        if not hasattr(self.__attr, name):
            raise AttributeError("%s object as no attribute '%s'" % (self.__attr.__class__.__name__, name))
        if callable(getattr(self.__attr, name)):
            return LDAPObjectAttributeCallableWrapper(self.__obj, getattr(self.__attr, name))
        return getattr(self.__attr, name)


class LDAPObjectAttributeCallableWrapper:  # pylint: disable=too-few-public-methods
    """
    Wrapper on LDAP object attribute's callable

    This wrapper add LDAP object reference as obj parameter on call (if accepted).
    """

    def __init__(self, obj, attr_callable):
        self.__obj = obj
        self.__callable = attr_callable

    def __call__(self, *args, **kwargs):
        if 'obj' in inspect.getfullargspec(self.__callable).args:
            kwargs['obj'] = self.__obj
        return self.__callable.__call__(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test on LDAP User object """

import datetime
import pytest

from eesyldap.exceptions import LDAPAttributeUnicityError
from eesyldap.exceptions import LDAPNotUniqueAttribute
from eesyldap.exceptions import LDAPRequiredAttribute
from eesyldap.exceptions import LDAPSearchHardLimitExceeded
from eesyldap.exceptions import LDAPSearchLimitExceeded
from eesyldap.search import LDAPSearch

from tests.exceptions import CantReuseOldPassword
from tests.exceptions import CantReuseSamePassword


def test_user_attributes_access(brenard):
    """ Check attributes values access """
    assert brenard.username == 'brenard'
    assert brenard.firstname == 'Benjamin'
    assert brenard.lastname == 'Renard'
    assert brenard.fullname == 'Benjamin Renard'
    assert brenard.fullname_normalized == 'Benjamin Renard'
    assert brenard.mail == 'brenard@easter-eggs.com'
    assert brenard.company_id == 'ee'
    assert sorted(brenard.service_ids) == ['admin', 'dev']
    assert isinstance(brenard.enabled, bool) and brenard.enabled
    assert isinstance(brenard.start_date, datetime.date) and brenard.start_date == datetime.date(year=2006, month=10, day=10)
    assert brenard.end_date is None
    assert isinstance(brenard.birth_date, datetime.date) and brenard.birth_date == datetime.date(year=1986, month=12, day=18)
    assert brenard.numid == 4
    assert brenard.bank_details == [dict(
        iban="FR76 3000 3034 3900 0504 9292 376",
        bic="SOGEFRPP",
        label="defaut"
    )]


def test_user_relations_access(brenard, Company, Service, Group):
    """ Test access to brenard relations """

    # Company
    company = brenard.company
    assert isinstance(company, Company) and company.dn == 'o=ee,ou=companies,o=eesyldap'

    # Managed company
    managed_company = brenard.managed_company
    assert isinstance(managed_company, Company)
    assert managed_company.dn == 'o=ee,ou=companies,o=eesyldap'
    assert managed_company.manager == brenard

    # Services
    services = brenard.services
    assert isinstance(services, dict)
    # Search is implicitly runned using len()
    assert len(services) == len(brenard.service_ids)
    for service_id, service in services.items():
        assert isinstance(service, Service)
        assert service.id == service_id
        assert service_id in brenard.service_ids
        assert service.manager == brenard

    # Managed services
    managed_services = brenard.managed_services
    assert isinstance(managed_services, LDAPSearch)
    assert managed_services.filterstr == "(&(objectClass=top)(objectClass=organizationalUnit)(objectClass=elService)(manager=uid=brenard,ou=people,o=eesyldap))"

    managed_services = dict(managed_services)
    assert len(managed_services) == 2
    for service_id, service in managed_services.items():
        assert isinstance(service, Service) and service_id in ['admin', 'dev']
        assert service.manager_dn == brenard.dn
        assert service.manager == brenard

    # Groups
    groups = brenard.groups
    assert isinstance(groups, LDAPSearch)
    # Search is implicitly runned using items()
    for group_name, group in groups.items():
        assert isinstance(group, Group)
        assert group.name == group_name
        assert brenard.dn in group.member_dns
        assert brenard.username in group.members
        assert brenard.username in group.members_by_username

    # Owned groups
    owned_groups = brenard.owned_groups
    assert isinstance(owned_groups, LDAPSearch) and owned_groups.filterstr == "(&(objectClass=top)(objectClass=elGroup)(owner=uid=brenard,ou=people,o=eesyldap))"
    assert len(owned_groups) == 1
    for group_name, group in owned_groups.items():
        assert isinstance(group, Group)
        assert group.name == group_name
        assert group.owner_dn == brenard.dn
        assert group.owner == brenard


def test_user_object_attribute_wrapper(brenard):
    """ Test object attributes wrapper """
    assert brenard.username == brenard.attrs.username.get_value()
    assert brenard.attrs.username.is_unique()


def test_user_save(brenard):
    """ Test changing user name """
    brenard.lastname = 'Fox'
    brenard.save()
    assert brenard.lastname == 'Fox'
    assert brenard.fullname == 'Benjamin Fox'


def test_user_update(brenard):
    """ Test multiple changes on user """
    changes = dict(
        username='b.renard',
        birth_date=datetime.date(1978, 12, 18),
        enabled=False,
    )
    brenard.update(changes)
    assert brenard.username == 'b.renard'
    assert brenard.birth_date == changes['birth_date']
    assert brenard.enabled == changes['enabled']


def test_user_password_change(brenard):
    """ Test change user password """
    new_password = "azerty"
    brenard.password = new_password
    brenard.save()

    assert brenard.password_last_change == datetime.date.today()
    assert brenard.attrs.password.verify(new_password)


def test_user_password_reuse_same(brenard):
    """ Test reuse same password is forbidden """
    new_password = "qwerty"
    brenard.password = new_password
    brenard.save()

    with pytest.raises(CantReuseSamePassword):
        brenard.password = new_password


def test_user_password_reuse_old(brenard):
    """ Test reuse an old password is forbidden """
    old_password = "azerty"
    new_password = "qwerty"
    brenard.password = old_password
    brenard.save()

    brenard.password = new_password
    brenard.save()

    with pytest.raises(CantReuseOldPassword):
        brenard.password = old_password


def test_user_unset_set_relation(brenard, ldap_client, Company):
    """ Test to unset a user relation """
    brenard.company = None
    assert brenard.company_id is None

    # Retrieve EE company
    ee = ldap_client.get(Company, id='ee')
    assert isinstance(ee, Company)
    assert ee.id == 'ee'

    # Set brenard company as ee
    brenard.company = ee
    assert brenard.company_id == 'ee'


def test_user_change_and_restore(brenard):
    """ Test to change user lastname and restore it locally """
    user_orig_lastname = brenard.lastname
    brenard.lastname = 'Fox'   # pylint: disable=attribute-defined-outside-init
    assert brenard.restore()
    assert brenard.lastname == user_orig_lastname
    assert not brenard.is_modified()


def test_user_change_and_refresh(brenard):
    """ Test to change user lastname and refresh it from LDAP after """
    user_orig_lastname = brenard.lastname
    brenard.lastname = 'Fox'   # pylint: disable=attribute-defined-outside-init
    assert brenard.refresh()
    assert brenard.lastname == user_orig_lastname


def test_user_search_on_username_numid(brenard, ldap_client, User):
    """ Test to search user on username and numid """
    search = ldap_client.search(User, username=brenard.username, numid=brenard.numid)
    assert isinstance(search, LDAPSearch)
    assert search.filterstr == '(&(objectClass=top)(objectClass=inetOrgPerson)(objectClass=elPeople)(elNumId=4)(uid=brenard))'
    assert len(search) == 1
    assert brenard == search[0]


def test_user_create(ldap_client, User):
    """ Test to create a user """
    attrs = dict(
        username='jdoe',
        firstname='John',
        lastname='Doe',
        fullname='John Doe',
        mail='jdoe@eesyldap.org',
        start_date=datetime.datetime.now().date(),
        birth_date=datetime.date(1982, 2, 22),
        numid=55,
    )
    user = ldap_client.new(User, attrs)
    assert isinstance(user, User)
    assert user.save()

    # Verify that we can now find this user on LDAP server
    search = ldap_client.search(User, username=attrs['username'])
    assert len(search) == 1
    assert user == search[0]


def test_user_delete(ldap_client, jdoe, User):
    """ Test to delete a user """
    assert jdoe.delete()

    verify = ldap_client.search(User, username='jdoe')
    assert not verify


def test_user_delete_with_relation(ldap_client, brenard):
    """ Test to delete a user with relation """
    with pytest.raises(LDAPRequiredAttribute):
        brenard.delete()


def test_user_search_by_username_in(ldap_client, User, brenard, jdoe):
    """ Test to search users by username value in a list """
    usernames = [brenard.username, jdoe.username]
    search = ldap_client.search(User, username=usernames)
    assert isinstance(search, LDAPSearch)
    assert search.filterstr == '(&(objectClass=top)(objectClass=inetOrgPerson)(objectClass=elPeople)(|(uid=brenard)(uid=jdoe)))'
    assert len(search) == 2
    assert brenard in search and jdoe in search


def test_user_search_sorted(ldap_client, User, brenard, jdoe):
    """ Test to search users and sort its """
    usernames = [brenard.username, jdoe.username]
    result = ldap_client.search(User, username=usernames, sort_by='username')
    assert result[1] == jdoe


def test_user_search_sorted_reverse(ldap_client, User, brenard, jdoe):
    """ Test to search users and sort its """
    usernames = [brenard.username, jdoe.username]
    result = ldap_client.search(User, username=usernames, sort_by='username', sort_reverse=True)
    assert result[1] == brenard


def test_user_search_with_key_attr(ldap_client, User, brenard):
    """ Test to search users with a key attribute """
    result = ldap_client.search(User, username=brenard.username, key_attr='numid')
    assert result[brenard.numid] == brenard


def test_user_search_with_bad_key_attr(ldap_client, User):
    """ Test to search users and sort its """
    with pytest.raises(LDAPNotUniqueAttribute):
        ldap_client.search(User, key_attr='lastname')


def test_user_access_search_result_by_index(ldap_client, User, brenard, jdoe):
    """ Test to search users and access result by index """
    usernames = [brenard.username, jdoe.username]
    result = ldap_client.search(User, username=usernames, sort_by='username')

    assert result[0] == brenard
    assert jdoe in result[1:2]
    assert result[-1] == jdoe


def test_user_search_with_limit(ldap_client, User, brenard, jdoe):
    """ Test to search users with limit """
    usernames = [brenard.username, jdoe.username]
    result = ldap_client.search(User, username=usernames, limit=1)
    with pytest.raises(LDAPSearchLimitExceeded):
        result.first()


def test_user_search_with_hardlimit(ldap_client, User, brenard, jdoe):
    """ Test to search users with hardlimit """
    usernames = [brenard.username, jdoe.username]
    result = ldap_client.search(User, username=usernames, hardlimit=1)
    with pytest.raises(LDAPSearchHardLimitExceeded):
        result.first()


def test_user_search_by_specify_filter_after_creation(ldap_client, User):
    """ Test to search users by specify options after search creation """
    search = ldap_client.search(User)
    search.filter(enabled=True)
    search.filter(filterstr='(elNumId=*)')
    assert search.filterstr == '(&(objectClass=top)(objectClass=inetOrgPerson)(objectClass=elPeople)(elEnabled=TRUE)(elNumId=*))'


def test_user_affect_not_unique_username(brenard, jdoe):
    """ Test to affect a not unique username to user """
    with pytest.raises(LDAPAttributeUnicityError):
        brenard.username = jdoe.username



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test commons things """

import datetime
import logging
import pytest

import unidecode

from eesyldap.attributes import Boolean, Date, Datetime, Integer, JSON, Password, ShadowLastChange, Text
from eesyldap.object import LDAPObject
from eesyldap.relations import RelationOnRemoteLinkAttribute, RelationOnLocalLinkAttribute
from eesyldap.unicity import UniqueInDirectory, UniqueInSameObjectType

from tests.exceptions import CantReuseOldPassword
from tests.exceptions import CantReuseSamePassword

log = logging.getLogger(__name__)


@pytest.fixture
def ldap_client():
    """ Start fake LDAP server and retreive correponding LDAPClient object """
    from ldap_test import LdapServer
    from eesyldap.client import LDAPClient

    base_entry = {
        'objectclass': 'organization',
        'dn': 'o=eesyldap',
        'attributes': {
            'o': 'eesyldap'
        }
    }

    entries = [
        # People
        {
            'objectclass': 'organizationalUnit',
            'dn': 'ou=people,o=eesyldap',
            'attributes': {
                'ou': 'people'
            }
        },
        {
            'objectclass': ['elPeople', 'inetOrgPerson', 'top'],
            'dn': 'uid=brenard,ou=people,o=eesyldap',
            'attributes': {
                'uid': 'brenard',
                'givenName': 'Benjamin',
                'sn': 'Renard',
                'displayName': 'Benjamin Renard',
                'cn': 'Benjamin Renard',
                'userPassword': '{SSHA}dXJlDxG6YAVex1YmfP7wJD6bceYrJcQY',
                'mail': 'brenard@easter-eggs.com',
                'o': 'ee',
                'ou': ['admin', 'dev'],
                'elEnabled': 'TRUE',
                'elStartDate': '20061010000000Z',
                'elLastLogin': '20190704103725Z',
                'elBirthDate': '19861218000000Z',
                'elNumId': '4',
                'elbankdetails': '{"iban": "FR76 3000 3034 3900 0504 9292 376", "bic": "SOGEFRPP", "label": "defaut"}',
            }
        },

        # Companies
        {
            'objectclass': 'organizationalUnit',
            'dn': 'ou=companies,o=eesyldap',
            'attributes': {
                'ou': 'companies'
            }
        },
        {
            'objectclass': ['top', 'organization', 'elCompany'],
            'dn': 'o=ee,ou=companies,o=eesyldap',
            'attributes': {
                'o': 'ee',
                'cn': 'Easter-eggs',
                'mail': 'info@easter-eggs.com',
                'elNumId': '1',
                'manager': 'uid=brenard,ou=people,o=eesyldap',
            }
        },

        # Services
        {
            'objectclass': 'organizationalUnit',
            'dn': 'ou=services,o=eesyldap',
            'attributes': {
                'ou': 'services'
            }
        },
        {
            'objectclass': ['top', 'organizationalUnit', 'elService'],
            'dn': 'ou=admin,ou=services,o=eesyldap',
            'attributes': {
                'ou': 'admin',
                'cn': 'Admin Sys',
                'o': 'ee',
                'mail': 'admin@easter-eggs.com',
                'elNumId': '2',
                'manager': 'uid=brenard,ou=people,o=eesyldap'
            }
        },
        {
            'objectclass': ['top', 'organizationalUnit', 'elService'],
            'dn': 'ou=dev,ou=services,o=eesyldap',
            'attributes': {
                'ou': 'dev',
                'cn': 'Dev',
                'o': 'ee',
                'mail': 'dev@easter-eggs.com',
                'elNumId': '3',
                'manager': 'uid=brenard,ou=people,o=eesyldap'
            }
        },

        # Groups
        {
            'objectclass': 'organizationalUnit',
            'dn': 'ou=groups,o=eesyldap',
            'attributes': {
                'ou': 'groups'
            }
        },
        {
            'objectclass': ['top', 'elGroup'],
            'dn': 'cn=ldapsuperadmins,ou=groups,o=eesyldap',
            'attributes': {
                'cn': 'ldapsuperadmins',
                'description': 'Super admins of the LDAP directory',
                'memberuid': 'brenard',
                'o': 'ee',
                'uniquemember': 'uid=brenard,ou=people,o=eesyldap',
                'owner': 'uid=brenard,ou=people,o=eesyldap',
            }
        },
    ]

    server = LdapServer({
        'base': base_entry,
        'entries': entries,
    })

    server.start()

    # Initialize client
    ldap_config = {
        'uri': 'ldap://localhost:%d' % server.config['port'],
        'base_dn': base_entry['dn'],
        'bind_dn': server.config['bind_dn'],
        'bind_credential': server.config['password'],
        'user.base_dn': 'ou=people,%s' % base_entry['dn'],
        'user.scope': 'one',
        'company.base_dn': 'ou=companies,%s' % base_entry['dn'],
        'company.scope': 'one',
        'service.base_dn': 'ou=services,%s' % base_entry['dn'],
        'service.scope': 'one',
        'group.base_dn': 'ou=groups,%s' % base_entry['dn'],
        'group.scope': 'one',
    }

    cache_params = dict(
        cache_enabled=False,
        cache_type='memory',
        cache_expire=600
    )

    yield LDAPClient(ldap_config, **cache_params)

    server.stop()


@pytest.fixture(scope="session")
def Service():
    class Service(LDAPObject):
        """ Test Service object class """

        _class = ['top', 'organizationalUnit', 'elService']
        _attrs = {
            'id': Text('ou', unicity=UniqueInSameObjectType(), required=True),
            'name': Text('cn', required=True),
            'company_id': Text('o', required=True),
            'description': Text('description'),
            'mail': Text('mail', unicity=UniqueInDirectory()),
            'numid': Integer('elNumId', unicity=UniqueInDirectory()),
            'manager_dn': Text('manager'),
        }
        _rdn_ldap_attr = 'ou'
    return Service


@pytest.fixture(scope="session")
def User(Service):
    class User(LDAPObject):
        """ Test User object class """

        _class = ['top', 'inetOrgPerson', 'elPeople']
        _attrs = {
            'username': Text('uid', unicity=UniqueInSameObjectType(), required=True),
            'password': Password('userPassword', hash_type='ssha'),
            'password_historic': Text('elUserPasswordHistoric', multiple=True),
            'password_last_change': ShadowLastChange('shadowLastChange'),
            'firstname': Text('givenName', required=True),
            'lastname': Text('sn', required=True),
            'fullname': Text('displayName', required=True),
            'fullname_normalized': Text('cn', required=True),
            'mail': Text('mail', unicity=UniqueInDirectory()),
            'company_id': Text('o'),
            'service_ids': Text('ou', multiple=True),
            'enabled': Boolean('elEnabled'),
            'start_date': Date('elStartDate'),
            'end_date': Date('elEndDate'),
            'last_login': Datetime('elLastLogin'),
            'birth_date': Date('elBirthDate', naive=True),
            'numid': Integer('elNumId', unicity=UniqueInDirectory()),
            'bank_details': JSON('elBankDetails', unicity=UniqueInDirectory(), multiple=True),
        }
        _rdn_ldap_attr = 'uid'
        _relations = {
            'services': RelationOnLocalLinkAttribute(Service, 'service_ids', link_attr_value='id', reverse_name='users', reverse_multiple=True),
            'managed_services': RelationOnRemoteLinkAttribute(Service, 'manager_dn', link_attr_value='dn', multiple=True, reverse_name='manager'),
        }

        def __setattr__(self, name, value):
            """ Set an attribute value """
            if name == 'password' and value:
                self._check_new_password(value)

            if LDAPObject.__setattr__(self, name, value):
                if name in ['firstname', 'lastname']:
                    self.fullname = "%s %s" % (self.firstname, self.lastname)  # pylint: disable=attribute-defined-outside-init
                if name == 'fullname':
                    self.fullname_normalized = unidecode.unidecode(value)  # pylint: disable=attribute-defined-outside-init
            return True

        def _check_new_password(self, new_password):
            """ Check new password """
            # Check new password against current password
            current_password = self.password
            log.debug("Current password = %s", current_password)
            if not self.password:
                log.debug('User have no password')
            elif self.attrs.password.verify(new_password, current_password):
                log.warning('Try to reuse the same password as current one')
                raise CantReuseSamePassword(self.attrs.password.ldap_name, new_password)

            # Check new password against historic
            historic = self.password_historic
            if self.attrs.password.is_hashed_password(new_password):
                log.warning('New password is currently hashed : could not check it against historic')
            else:
                log.debug('Current password historic = %s', historic)
                for hashed_password in historic:
                    if self.attrs.password.verify(new_password, hashed_password):
                        log.warning('Try to reuse an old password')
                        raise CantReuseOldPassword(self.attrs.password.ldap_name, new_password)
                    log.debug('Password "%s" does not match with old one "%s"', new_password, hashed_password)

            log.debug('New user password is valid')

            # Add current password to historic
            if current_password:
                if len(historic) >= 3:
                    historic = historic[(len(historic) - 4):]
                if not self.attrs.password.is_hashed_password(current_password):
                    log.info('Current password is clear : hash it before inclusion in historic')
                    current_password = self.attrs.password.hash(current_password)
                if current_password not in historic:
                    historic.append(current_password)
                log.debug('New user password historic = %s', historic)
                self.password_historic = historic   # pylint: disable=attribute-defined-outside-init
            else:
                log.debug("Current user password not available : can't add it in historic")

            # Adjust password_last_change
            self.password_last_change = datetime.date.today()   # pylint: disable=attribute-defined-outside-init

    # Add reverse relations of User objects to Services
    User.add_reverse_relations()

    return User


@pytest.fixture(scope="session")
def Company(User, Service):
    class Company(LDAPObject):
        """ Test Company object class """

        _class = ['top', 'organization', 'elCompany']
        _attrs = {
            'id': Text('o', unicity=UniqueInSameObjectType(), required=True),
            'name': Text('cn', unicity=UniqueInSameObjectType(), required=True),
            'description': Text('description'),
            'mail': Text('mail', unicity=UniqueInDirectory()),
            'numid': Integer('elNumId', unicity=UniqueInDirectory()),
            'manager_dn': Text('manager', required=True),
        }
        _rdn_ldap_attr = 'o'

        _relations = {
            'manager': RelationOnLocalLinkAttribute(User, 'manager_dn', link_attr_value='dn', reverse_name='managed_company', reverse_multiple=False),
            'users': RelationOnRemoteLinkAttribute(User, 'company_id', link_attr_value='id', reverse_name='company'),
            'services': RelationOnRemoteLinkAttribute(Service, 'company_id', link_attr_value='id', reverse_name='company'),
        }

    # Add reverse relations of Company objects
    Company.add_reverse_relations()

    return Company


@pytest.fixture(scope="session")
def Group(User, Service, Company):
    class Group(LDAPObject):
        """ Test Group object class """
        _class = ['top', 'elGroup']
        _attrs = {
            'name': Text('cn', unicity=UniqueInSameObjectType(), required=True),
            'description': Text('description', required=True),
            'member_usernames': Text('memberUid', multiple=True),
            'member_dns': Text('uniqueMember', multiple=True),
            'company_id': Text('o'),
            'service_ids': Text('ou', multiple=True),
            'owner_dn': Text('owner'),
        }

        _rdn_attr = 'name'

        _relations = {
            'owner': RelationOnLocalLinkAttribute(User, 'owner_dn', link_attr_value='dn', reverse_name='owned_groups'),
            'members': RelationOnLocalLinkAttribute(User, 'member_dns', link_attr_value='dn', reverse_name='groups', reverse_multiple=True),
            'members_by_username': RelationOnLocalLinkAttribute(User, 'member_usernames', link_attr_value='username', reverse_name='groups_by_username', reverse_multiple=True),
            'company': RelationOnLocalLinkAttribute(Company, 'company_id', link_attr_value='id', reverse_name='groups'),
            'services': RelationOnLocalLinkAttribute(Service, 'service_ids', link_attr_value='id', reverse_name='groups'),
        }

        def __setattr__(self, name, value):
            """ Set an attribute value """
            if LDAPObject.__setattr__(self, name, value):
                if name == 'members':
                    self.members_by_username = dict(    # pylint: disable=attribute-defined-outside-init
                        (member.username, member)
                        for member_dn, member in value.items()
                    )
            return True

    # Add reverse relations of Group objects
    Group.add_reverse_relations()

    return Group


@pytest.fixture
def brenard(ldap_client, User):
    """ Retreive brenard user from LDAP """
    user = ldap_client.get(User, username='brenard')
    assert isinstance(user, User)
    assert user.username == 'brenard'

    return user


@pytest.fixture
def jdoe(ldap_client, User):
    """ Create jdoe user in LDAP """
    attrs = dict(
        username='jdoe',
        firstname='John',
        lastname='Doe',
        fullname='John Doe',
        mail='jdoe@eesyldap.org',
        start_date=datetime.datetime.now().date(),
        birth_date=datetime.date(1982, 2, 22),
        numid=55,
    )
    user = ldap_client.new(User, attrs)
    assert isinstance(user, User)
    assert user.save()

    return user


@pytest.fixture
def ee(ldap_client, Company):
    """ Retreive ee company from LDAP """
    company = ldap_client.get(Company, id='ee')
    assert isinstance(company, Company)
    assert company.id == 'ee'

    return company


@pytest.fixture
def dev(ldap_client, Service):
    """ Retreive dev service from LDAP """
    service = ldap_client.get(Service, id='dev')
    assert isinstance(service, Service)
    assert service.id == 'dev'

    return service


@pytest.fixture
def admin(ldap_client, Service):
    """ Retreive admin service from LDAP """
    service = ldap_client.get(Service, id='admin')
    assert isinstance(service, Service)
    assert service.id == 'admin'

    return service


@pytest.fixture
def ldapsuperadmins(ldap_client, Group):
    """ Retreive admin service from LDAP """
    group = ldap_client.get(Group, name='ldapsuperadmins')
    assert isinstance(group, Group)
    assert group.name == 'ldapsuperadmins'

    return group



# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Tests on LDAP filters helpers """

import pytest

from eesyldap.exceptions import LDAPInvalidFilterLogicalOperator
from eesyldap.exceptions import LDAPInvalidFilterMatchRule
from eesyldap.filters import combine
from eesyldap.filters import compose


def test_compose_equals():
    assert compose('fakeAttr', 'equals', 'value') == '(fakeAttr=value)'
    assert compose('fakeAttr', '=', 'value') == '(fakeAttr=value)'
    assert compose('fakeAttr', '==', 'value') == '(fakeAttr=value)'


def test_compose_not_equals():
    assert compose('fakeAttr', 'not_equals', 'value') == '(!(fakeAttr=value))'


def test_compose_equals_none():
    with pytest.raises(AssertionError):
        compose('fakeAttr', 'equals')


def test_compose_begins():
    assert compose('fakeAttr', 'begins', 'value') == '(fakeAttr=value*)'


def test_compose_not_begins():
    assert compose('fakeAttr', 'not_begins', 'value') == '(!(fakeAttr=value*))'


def test_compose_ends():
    assert compose('fakeAttr', 'ends', 'value') == '(fakeAttr=*value)'


def test_compose_not_ends():
    assert compose('fakeAttr', 'not_ends', 'value') == '(!(fakeAttr=*value))'


def test_compose_contains():
    assert compose('fakeAttr', 'contains', 'value') == '(fakeAttr=*value*)'


def test_compose_not_contains():
    assert compose('fakeAttr', 'not_contains', 'value') == '(!(fakeAttr=*value*))'


def test_compose_greater():
    assert compose('fakeAttr', 'greater', '1') == '(fakeAttr>1)'
    assert compose('fakeAttr', '>', '1') == '(fakeAttr>1)'


def test_compose_not_greater():
    assert compose('fakeAttr', 'not_greater', '1') == '(!(fakeAttr>1))'


def test_compose_less():
    assert compose('fakeAttr', 'less', '1') == '(fakeAttr<1)'
    assert compose('fakeAttr', '<', '1') == '(fakeAttr<1)'


def test_compose_not_less():
    assert compose('fakeAttr', 'not_less', '1') == '(!(fakeAttr<1))'


def test_compose_greater_or_equal():
    assert compose('fakeAttr', 'greater_or_equal', '1') == '(fakeAttr>=1)'
    assert compose('fakeAttr', '>=', '1') == '(fakeAttr>=1)'


def test_compose_not_greater_or_equal():
    assert compose('fakeAttr', 'not_greater_or_equal', '1') == '(!(fakeAttr>=1))'


def test_compose_less_or_equal():
    assert compose('fakeAttr', 'less_or_equal', '1') == '(fakeAttr<=1)'
    assert compose('fakeAttr', '<=', '1') == '(fakeAttr<=1)'


def test_compose_not_less_or_equal():
    assert compose('fakeAttr', 'not_less_or_equal', '1') == '(!(fakeAttr<=1))'


def test_compose_approx():
    assert compose('fakeAttr', 'approx', 'value') == '(fakeAttr~=value)'
    assert compose('fakeAttr', '~=', 'value') == '(fakeAttr~=value)'


def test_compose_not_approx():
    assert compose('fakeAttr', 'not_approx', 'value') == '(!(fakeAttr~=value))'


def test_compose_present():
    assert compose('fakeAttr', 'present') == '(fakeAttr=*)'
    assert compose('fakeAttr', 'any') == '(fakeAttr=*)'


def test_compose_not_present():
    assert compose('fakeAttr', 'not_present') == '(!(fakeAttr=*))'
    assert compose('fakeAttr', 'not_any') == '(!(fakeAttr=*))'


def test_compose_invalid_match():
    with pytest.raises(LDAPInvalidFilterMatchRule):
        assert compose('fakeAttr', 'error', 'value')


def test_compose_dont_escape_value():
    values = '*()\\\u0000ée'
    for char in values:
        assert compose('fakeAttr', '==', char, escape=False) == '(fakeAttr=%s)' % char


def test_compose_escape_values():
    values = [
        ('*', '\\2a'),
        ('(', '\\28'),
        (')', '\\29'),
        ('\\', '\\5c'),
        ('\u0000', '\\00'),
        ('é', 'é'),
        ('e', 'e')
    ]
    for value, escape_value in values:
        assert compose('fakeAttr', '==', value, escape_mode=0) == '(fakeAttr=%s)' % escape_value


def test_compose_escape_values_nonascii_chars():
    values = [
        ('*', '\\2a'),
        ('(', '\\28'),
        (')', '\\29'),
        ('\\', '\\5c'),
        ('\u0000', '\\00'),
        ('é', '\\e9'),
        ('e', 'e')
    ]
    for value, escape_value in values:
        assert compose('fakeAttr', '==', value, escape_mode=1) == '(fakeAttr=%s)' % escape_value


def test_compose_escape_values_all_chars():
    values = [
        ('*', '\\2a'),
        ('(', '\\28'),
        (')', '\\29'),
        ('\\', '\\5c'),
        ('\u0000', '\\00'),
        ('é', '\\e9'),
        ('e', '\\65')
    ]
    for value, escape_value in values:
        assert compose('fakeAttr', '==', value, escape_mode=2) == '(fakeAttr=%s)' % escape_value


def test_combine_and():
    assert combine('and', '(fakeAttr1=test1)', '(fakeAttr2=test2)') == '(&(fakeAttr1=test1)(fakeAttr2=test2))'
    assert combine('&', '(fakeAttr1=test1)', '(fakeAttr2=test2)') == '(&(fakeAttr1=test1)(fakeAttr2=test2))'
    with pytest.raises(AssertionError):
        combine('and', '(fakeAttr1=test1)')
    with pytest.raises(AssertionError):
        combine('&', '(fakeAttr1=test1)')


def test_combine_or():
    assert combine('or', '(fakeAttr1=test1)', '(fakeAttr2=test2)') == '(|(fakeAttr1=test1)(fakeAttr2=test2))'
    assert combine('|', '(fakeAttr1=test1)', '(fakeAttr2=test2)') == '(|(fakeAttr1=test1)(fakeAttr2=test2))'
    with pytest.raises(AssertionError):
        combine('or', '(fakeAttr1=test1)')
    with pytest.raises(AssertionError):
        combine('|', '(fakeAttr1=test1)')


def test_combine_not():
    assert combine('not', '(fakeAttr1=test1)') == '(!(fakeAttr1=test1))'
    assert combine('!', '(fakeAttr1=test1)') == '(!(fakeAttr1=test1))'
    with pytest.raises(AssertionError):
        combine('not', '(fakeAttr1=test1)', '(fakeAttr2=test2)')
    with pytest.raises(AssertionError):
        combine('!', '(fakeAttr1=test1)', '(fakeAttr2=test2)')


def test_combine_invalid_log_op():
    with pytest.raises(LDAPInvalidFilterLogicalOperator):
        assert combine('error', '(fakeAttr1=test1)', '(fakeAttr2=test2)')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=too-few-public-methods

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Attributes unicity checks """

import logging

from eesyldap.exceptions import LDAPAttributeUnicityError
from eesyldap.filters import combine as combine_ldap_filter
from eesyldap.filters import compose as compose_ldap_filter

log = logging.getLogger(__name__)


class BaseUnicityCheck:
    """
    Base class for LDAP Attributes unicity checks

    This class is not supposed to be used as unicity checker (see its chidren).
    """

    def __init__(self, **kwargs):
        """
        Constructor

        This method handle keyword arguments and define corresponding object attributes (if exists).
        """
        for arg, value in kwargs.items():
            if hasattr(self, arg):
                setattr(self, arg, value)
            else:
                raise AttributeError("Invalid keyword argument %s pass to %s constructor" % (arg, self.__class__))

    def is_unique(self, client, obj, attr, value):
        """
        Check attribute value unicity

        :param  client:     The LDAPClient object
        :param  obj:        The LDAPObject object
        :param  attr:       The attribute name to check
        :param  value:      The attribute value to check
        """
        raise NotImplementedError("%s.%s" % (self.__repr__(), 'is_unique'))

    def __repr__(self):
        """ Compute and return the “official” string representation of the Unicity Check """
        return "<%s>" % type(self).__name__


class UniqueInDirectory(BaseUnicityCheck):  # pylint: disable=too-few-public-methods
    """
    LDAP attributes unicity check on all directory objects

    This unicity check that the attribute value is unique on all of the LDAP directory objects.

    By default, deplicated values are searched on all objects in the directory with the same attribute.
    You can configured other attibutes that could stores duplicated values by using the other_attributes
    constructor parameter.
    """

    # Also check unicity on other attributes
    other_attributes = []

    # Do not raise error if the duplicated value is own by the same object
    excepted_itself = True

    def is_unique(self, client, obj, attr, value):
        """
        Check attribute value unicity

        :param  client:     The LDAPClient object
        :param  obj:        The LDAPObject object
        :param  attr:       The attribute name to check
        :param  value:      The attribute value to check
        """
        return self._is_unique(client, obj, attr, value)

    def _is_unique(self, client, obj, attr, value, base_dn=None):
        """
        Real (and overriddable) implementation of the attribute value unicity check

        :param  client:     The LDAPClient object
        :param  obj:        The LDAPObject object
        :param  attr:       The attribute name to check
        :param  value:      The attribute value to check
        :param  base_dn:    The base DN of the LDAP search of duplicated values (optional, defaut : LDAPClient directory configured base DN)
        """
        filterstr, attrs = self._compose_filterstr_and_attrlist(client, obj, attr, value)
        result = client.raw_search(filterstr, attrlist=attrs, base_dn=base_dn)
        if isinstance(result, dict):
            if result and self.excepted_itself and obj.dn in result:
                del result[obj.dn]
            if not result:
                return True
            raise LDAPAttributeUnicityError(attr, value, result.keys())
        log.error('An unexpected error occured checking the unicity of the value "%s" of the attribute %s of the object %s with the check %s', value, attr.ldap_name, obj, self)
        return False

    def _compose_filterstr_and_attrlist(self, client, obj, attr, value):  # pylint: disable=unused-argument
        """
        Compose the filterstr and the attributes list of the LDAP search of duplicated values

        :param  client:     The LDAPClient object
        :param  obj:        The LDAPObject object
        :param  attr:       The attribute name to check
        :param  value:      The attribute value to check
        """
        filterstr = compose_ldap_filter(attr.ldap_name, 'equals', value.decode('utf8'))
        attrs = [attr.ldap_name]
        if self.other_attributes:
            assert isinstance(self.other_attributes, list)
            filterstrs = [filterstr]
            for other_attr in self.other_attributes:
                filterstrs.append(compose_ldap_filter(other_attr, 'equals', value.decode('utf8')))
                attrs.append(other_attr)
            filterstr = combine_ldap_filter('or', *filterstrs)
        return (filterstr, attrs)


class UniqueInSameObjectType(UniqueInDirectory):
    """
    LDAP attributes unicity check on all directory objects of the same type

    This unicity check that the attribute value is unique on all of the same type LDAP directory objects.

    This class derive from UniqueInDirectory and override the parameter of the LDAP search of duplicated
    values. Duplicated values are searched in all objects of the same type and in the same attribute.
    As with UniqueInDirectory check, you can configured other attibutes that could stores duplicated
    values by using the other_attributes constructor parameter.
    """

    def is_unique(self, client, obj, attr, value):
        """
        Check attribute value unicity

        :param  client:     The LDAPClient object
        :param  obj:        The LDAPObject object
        :param  attr:       The attribute name to check
        :param  value:      The attribute value to check
        """
        return self._is_unique(client, obj, attr, value, base_dn=client.get_object_base_dn(obj.__class__))

    def _compose_filterstr_and_attrlist(self, client, obj, attr, value):  # pylint: disable=unused-argument
        """
        Compose the filterstr and the attributes list of the LDAP search of duplicated values

        :param  client:     The LDAPClient object
        :param  obj:        The LDAPObject object
        :param  attr:       The attribute name to check
        :param  value:      The attribute value to check
        """
        filterstr, attrs = super()._compose_filterstr_and_attrlist(client, obj, attr, value)
        filterstr = obj.get_filter(filterstr=filterstr)
        return (filterstr, attrs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP search """

import copy
from collections import OrderedDict
import logging

from eesyldap.exceptions import LDAPInvalidAttributeName
from eesyldap.exceptions import LDAPNotUniqueAttribute
from eesyldap.exceptions import LDAPSearchHardLimitExceeded
from eesyldap.exceptions import LDAPSearchLimitExceeded
from eesyldap.filters import combine as combine_ldap_filter

log = logging.getLogger(__name__)


class LDAPSearch:   # pylint: disable=too-many-instance-attributes
    """
    LDAP search

    This class abstract an LDAP search as a python object emulating a list object (or dict object if key_attr is provided).
    All search parameters could be provied using constructor parameters. You also could use the filter method to specify a filter
    or the sort method to specify sort parameters.

    :param  client:         The LDAPClient object
    :param  obj_type:       The object type to search
    :param  base_dn:        The base DN of the search (optional)
    :param  filterstr:      An additionnal specific LDAP filter string (optional)
    :param  scope:          The scope of the search (optional)
    :param  limit:          Specific the maximum number of object you want to retreived (optional, default: no limit)
                            If the limit is exceeded, an LDAPSearchLimitExceeded exception will be raised.
    :param  hardlimit:      Specific the maximum number of object you want to retreived (optional, default: no limit)
                            This limit is handled by LDAP server and potentially included references. Prefer limit
                            parameter usage to avoid counting references. If the limit is exceeded, an
                            LDAPSearchHardLimitExceeded exception will be raised.
    :param  sort_by:        Sort resulting objects list by an specific attribute (optional)
    :param  sort_reverse:   If true, the sorted resulting objects is reversed (optional, only use if sort_by is defined)
    :param  sort_limit:     If the number of objects to sort if upper to the specify limit, do not sort the result (optional,
                            default: no limit).
    :param  refresh:        Set to True to be sure that result is not retreive from cache (optional, default: False)
    :param  key_attr:       Set witch attribute while be used as key (optional, conflicting with sorted search)
    :param  filters:        Dict of filters to search on (optional)
    """

    _client = None
    _obj_type = None
    _filterstrs = None
    _key_attr = None
    _refresh = None

    # Sort parameters
    _sort_by = None
    _sort_reverse = False
    _sort_limit = None

    # Limit parameter
    _limit = None

    # Hard-limit parameter
    _hardlimit = None

    # Raw LDAP search result
    __result = None

    # Objects DN list (sorted if sort is enabled)
    __obj_dns = None

    # Objects key to DN mapping (if kew_attr is defined)
    __obj_key_to_dn = None

    # Objects cache
    __objs = {}

    def __init__(self, client, obj_type, base_dn=None, filterstr=None, scope=None, limit=None, hardlimit=None,
                 sort_by=None, sort_reverse=False, sort_limit=None, key_attr=None, refresh=False, **filters):
        from eesyldap.client import LDAPClient
        assert isinstance(client, LDAPClient), "obj_type argument must be an LDAPClient object (not %s)" % client
        self._client = client
        self._obj_type = obj_type
        self._base_dn = base_dn or client.get_object_base_dn(obj_type)
        self._scope = scope or client.get_object_scope(obj_type)
        self._refresh = bool(refresh)

        # Handle filters
        self._filterstrs = []
        if filterstr or filters:
            self.filter(filterstr=filterstr, **filters)

        # Handle limit/hardlimit parameters
        assert limit is None or isinstance(limit, int)
        self._limit = limit
        assert hardlimit is None or isinstance(hardlimit, int)
        self._hardlimit = hardlimit

        # Handle key attribute
        if key_attr:
            if obj_type.exists(key_attr):
                # Check attribute is unique
                if not obj_type.is_unique(key_attr):
                    raise LDAPNotUniqueAttribute(key_attr)
            elif key_attr != 'dn':
                raise LDAPInvalidAttributeName(key_attr)
            self._key_attr = key_attr

        # Handle sort parameters
        if sort_by:
            self.sort(sort_by, reverse=sort_reverse, limit=sort_limit)

    def sort(self, sort_by, reverse=False, limit=None):
        """
        Set search sort parameters

        :param  sort_by:    Sort resulting objects list by an specific attribute
        :param  reverse:    If true, the sorted resulting objects is reversed (optional, default: False)
        :param  limit:      If the number of objects to sort if upper to the specify limit, do not sort the
                            result (optional, default: no limit).
        """
        assert self._key_attr is None, "You could not sort search with key attribute defined"
        if sort_by != 'dn' and not self._obj_type.exists(sort_by):
            raise LDAPNotUniqueAttribute(sort_by)

        self._sort_by = sort_by
        self._sort_reverse = bool(reverse)
        if limit is not None:
            assert isinstance(limit, int)
            self._sort_limit = limit
        else:
            self._sort_limit = self._client.get_config('sort.limit', obj_type=self._obj_type)
        self._clear_sort_result()

    def filter(self, *filterstrs, filterstr=None, log_op=None, **filters):
        """
        Set additionnal search filters

        :param  filterstr:          LDAP filter string
        :param  filterstrs:         List of LDAP filter strings
        :param  filters:            Dict of filter attributes
        :param  log_op:             Logical operator use to combine provided fitler with existing
                                    ones (optional, default: and)
        """
        log_op = 'and' if log_op is None else log_op
        assert filterstr or filterstrs or filters, "You must provided at least one filter"
        if filterstr:
            self._filterstrs.append(filterstr)
        self._filterstrs.extend(filterstrs)
        for (attr_name, attr_value) in filters.items():
            self._filterstrs.append(self._obj_type.get_attribute_filter(attr_name, attr_value))

        # If log_op is not and/&, we need to combine filter strings. Otherwise, it will be done before search
        if log_op != 'and' and log_op != '&' and len(self._filterstrs) > 1:
            # Sort filter strings before combined it to maximum the chance to retreived
            # always the same filter string (usefull for caching key)
            self._filterstrs = [combine_ldap_filter(log_op, *sorted(self._filterstrs))]
        self._clear_result()

    def _get_filterstr(self):
        """ Return the LDAP filter string of the search """
        # Sort filter strings before combined it to maximum the chance to retreived
        # always the same filter string (usefull for caching key)
        return self._obj_type.get_filter(*sorted(self._filterstrs))

    def __getattr__(self, name):
        """ Get specific attribute value """
        if name in ['obj_type', 'client']:
            return getattr(self, "_" + name)
        if name in ['base_dn', 'scope', 'key_attr', 'refresh', 'sort_by', 'sort_reverse', 'sort_limit', 'limit', 'hardlimit']:
            return copy.copy(getattr(self, "_" + name))
        if name == 'filterstr':
            return self._get_filterstr()
        raise AttributeError("%s object has no %s attribute" % (self.__class__.__name__, name))

    def _clear_result(self):
        """ Clear result """
        self.__result = None
        self.__objs = {}
        self.__obj_key_to_dn = None
        self._clear_sort_result()

    def _clear_sort_result(self):
        """ Clear sort result """
        self.__obj_dns = None

    def _execute(self, refresh=False):
        """
        Execute prepared search

        :param  refresh:    Force refresh mode (do not retreived informations from cache or local search result)
        """
        # Check if search is already exeeded (and not refresh mode)
        if isinstance(self.__result, OrderedDict) and not refresh:
            return True

        # Clear result
        self._clear_result()

        # Check if result is in cache (if enabled and not refresh mode)
        self.__result = self._client.get_cached_object_search(
            self._obj_type,
            self.filterstr,
            self._base_dn,
            self._scope,
            refresh=refresh or self._refresh
        )

        if isinstance(self.__result, OrderedDict):
            # Ensure hard limit is not exceeded
            if self._hardlimit and len(self.__result) > self._hardlimit:
                raise LDAPSearchHardLimitExceeded(self._hardlimit)
            # Ensure limit is not exceeded
            if self._limit and len(self.__result) > self._limit:
                raise LDAPSearchLimitExceeded(self._limit)
        else:
            # Search matching objects in LDAP
            log.debug('Execute LDAP search')
            filterstr = self._get_filterstr()
            self.__result = self._client.raw_search(
                filterstr,
                base_dn=self._base_dn,
                scope=self._scope,
                attrlist=self._obj_type.ldap_attr_names(),
                limit=self._limit,
                hardlimit=self._hardlimit
            )

            # Ensure search result is an OrderedDict
            if not isinstance(self.__result, OrderedDict):
                # Otherwise, return the raw result of the search
                return self.__result

            if self.__result:
                log.debug('%s matching %s object(s) found in LDAP', len(self.__result), self._obj_type.__name__)
                # Add search result in cache (only if not empty)
                self._client.add_cached_object_search(
                    self._obj_type,
                    filterstr,
                    self._base_dn,
                    self._scope,
                    self.__result
                )
            else:
                log.debug('No %s matching object found in LDAP', self._obj_type.__name__)

        # If key attribute is defined, generate key to DN mapping
        if self._key_attr:
            self.__obj_key_to_dn = dict()
            for obj_dn, obj_attrs_values in self.__result.items():
                key = self._obj_type.get_python_attribute_value(self._key_attr, obj_attrs_values)
                assert key not in self.__obj_key_to_dn, 'Duplicated object key value "%s"' % key
                self.__obj_key_to_dn[key] = obj_dn
        return True

    def _sort(self):
        """ Sort search result (if enabled and sort limit is not exceeded) """

        # Check if result is already sorted
        if isinstance(self.__obj_dns, list):
            return True

        # Check if search is not executed
        if self.__result is None:
            return False

        if self._sort_by:
            # Check if sorted search result is in the cache
            self.__obj_dns = self._client.get_cached_object_sorted_search(
                self._obj_type,
                self.filterstr,
                self._base_dn,
                self._scope,
                self._sort_by,
                self._sort_reverse,
                refresh=self._refresh
            )

            # Otherwise, sort the search result
            if not self.__obj_dns:
                if self._sort_limit:
                    if len(self.__result) > self._sort_limit:
                        log.debug('Sort limit exceeded (%s / %s): do not sort result by %s', len(self.__result), self._sort_limit, self._sort_by)
                        self.__obj_dns = list(self.__result)
                    else:
                        log.debug('Sort limit not exceeded (%s / %s): sort result by %s', len(self.__result), self._sort_limit, self._sort_by)
                else:
                    log.debug('Sort search result by %s', self._sort_by)

                if not self.__obj_dns:
                    # Define method use as sorted() key parameter to sort search result
                    def _sort_by(obj_dn):
                        if self._sort_by == 'dn':
                            return obj_dn
                        return self._obj_type.get_python_attribute_value(self._sort_by, self.__result[obj_dn], sortable=True)
                    self.__obj_dns = list(self.__result)
                    self.__obj_dns = sorted(self.__obj_dns, key=_sort_by, reverse=self._sort_reverse)
                    self._client.add_cached_object_sorted_search(
                        self._obj_type,
                        self.filterstr,
                        self._base_dn,
                        self._scope,
                        self._sort_by,
                        self._sort_reverse,
                        self.__obj_dns
                    )
        else:
            self.__obj_dns = list(self.__result)
        return True

    def get_object_by_dn(self, obj_dn):
        """ Retreive an object of the search result by it DN """
        # Ensure search is runned and object is in the result
        if not self._execute():
            return None
        assert obj_dn in self.__result

        # Create object if not exists
        if obj_dn not in self.__objs:
            self.__objs[obj_dn] = self._obj_type(self._client, dn=obj_dn, attrs_values=self.__result[obj_dn])

        return self.__objs[obj_dn]

    def contains(self, obj_dn):
        """ Check if an object is contained in search result by its DN """
        # Ensure search is runned and object is in the result
        if not self._execute():
            return None
        return obj_dn in self.__result

    def __len__(self):
        """ Get search result length """
        if not self._execute():
            return 0
        return len(self.__result)

    def __bool__(self):
        """ Check is result is empty or not """
        if not self._execute():
            return False
        return bool(self.__result)

    def __getitem__(self, key):
        """ Get an item of the result """
        log.debug('%s.__getitem__(%s)', self, key)
        if not self._execute() or not self._sort():
            return None
        if self._key_attr:
            # Behave like a dict
            if key not in self.__obj_key_to_dn:
                raise IndexError(key)
            obj_dn = self.__obj_key_to_dn[key]
        else:
            # Behave like a list
            if isinstance(key, int):
                if key < 0:
                    # Handle negative indices
                    key += len(self)

                if key < 0 or key >= len(self):
                    # Invalid indice
                    raise IndexError('The index ({0}) is out of range.'.format(key))

                obj_dn = self.__obj_dns[key]
            elif isinstance(key, slice):
                return [self[i] for i in range(*key.indices(len(self)))]
            else:
                raise TypeError('Invalid argument type.')

        return self.get_object_by_dn(obj_dn)

    def __iter__(self):
        """ Iter on search result """
        if self._execute() and self._sort():
            for obj_dn in self.__obj_dns:
                if self._key_attr:
                    # Behave like a dict : return the key
                    if self._key_attr == 'dn':
                        yield obj_dn
                    else:
                        yield self._obj_type.get_python_attribute_value(self._key_attr, self.__result[obj_dn])
                else:
                    # Behave like a list : return the object
                    yield self.get_object_by_dn(obj_dn)

    def first(self):
        """ Retreive first object """
        if not self._execute() or not self._sort():
            return None
        obj_dn = self.__obj_dns[0]
        return self.get_object_by_dn(obj_dn)

    def keys(self):
        """ Retreive results keys """
        if not self._execute() or not self._sort():
            return None
        if self._key_attr:
            # Behave like a dict : return list of the keys
            return list(self)
        # List do not implement keys() method : return the index range
        return range(len(self))

    def items(self):
        """ Retreived search result items (as a dict) """
        # Behave like a dict : return list of tuple(key, obj)
        return [(key, self[key]) for key in self.keys()]

    def __repr__(self):
        """ Compute and return the “official” string representation of a LDAPSearch object """
        return '<LDAPSearch on %s objects with filter "%s" on base DN %s (scope=%s). State = %s>' % (
            self._obj_type.__name__,
            self.filterstr,
            self._base_dn,
            self._scope,
            ("Executed with %s result(s)" % len(self)) if isinstance(self.__result, OrderedDict) else 'Not executed'
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Test on LDAP company object """

from eesyldap.search import LDAPSearch


def test_company_attributes_access(ee):
    """ Test access to company's attributes """
    assert ee.id == 'ee'
    assert ee.name == 'Easter-eggs'
    assert ee.mail == 'info@easter-eggs.com'
    assert ee.numid == 1
    assert ee.manager_dn == 'uid=brenard,ou=people,o=eesyldap'


def test_company_relations_access(ee, brenard, Group, Service):
    """ Test access to company relations """

    # Groups
    groups = ee.groups
    assert 'ldapsuperadmins' in groups
    assert isinstance(groups['ldapsuperadmins'], Group)
    assert groups['ldapsuperadmins'].dn == 'cn=ldapsuperadmins,ou=groups,o=eesyldap'

    # Manager
    assert ee.manager == brenard

    # Users
    users = ee.users
    assert isinstance(users, LDAPSearch) and users.filterstr == '(&(objectClass=top)(objectClass=inetOrgPerson)(objectClass=elPeople)(o=ee))'
    assert len(users) == 1
    assert brenard.username in users
    assert users[brenard.username] == brenard

    # Services
    services = ee.services
    assert isinstance(services, LDAPSearch) and services.filterstr == '(&(objectClass=top)(objectClass=organizationalUnit)(objectClass=elService)(o=ee))'
    assert len(services) == 2
    for service_id in services:
        assert service_id in ['admin', 'dev']
        assert isinstance(services[service_id], Service)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Exceptions """

from ldap import SIZELIMIT_EXCEEDED     # pylint: disable=no-name-in-module


class LDAPException(Exception):
    """ That is the base exception class for all the other exceptions provided by this module. """


class LDAPInvalidAttributeName(AttributeError, LDAPException):
    """ Raised when an invalid attribute is call on an LDAPObject """

    def __init__(self, attr_name):
        self.attr_name = attr_name
        super().__init__("Invalid attribute name '%s'" % attr_name)


class LDAPInvalidAttributeValue(TypeError, LDAPException):
    """ Raised when an invalid value is affected to an attribute on an LDAPObject """

    def __init__(self, attr_name, attr_value, msg=None):
        self.attr_name = attr_name
        self.attr_value = attr_value
        super().__init__(msg or "Invalid value affected to attribute '%s'" % attr_name)


class LDAPNotUniqueAttribute(TypeError, LDAPException):
    """ Raised when an attribute is call as an LDAPObject identifier
    but is not unique."""

    def __init__(self, attr_name):
        self.attr_name = attr_name
        super().__init__("Attribute %s is not unique" % attr_name)


class LDAPAttributeUnicityError(LDAPInvalidAttributeValue):
    """ Raised when try to set a value to an LDAP attribute and when this
    value is already affected to another object(s)."""

    def __init__(self, attr_name, attr_value, other_objects_dn):
        self.other_objects_dn = other_objects_dn
        super().__init__(attr_name, attr_value, msg="The value '%s' of the attribute %s is not unique. This value is already affected to the following LDAP object(s) : %s" % (attr_value.decode('utf-8'), attr_name, ', '.join(other_objects_dn)))


class LDAPRequiredAttribute(LDAPInvalidAttributeValue):
    """ Raised when an attribute is set to None while is required. """

    def __init__(self, attr_name):
        super().__init__(attr_name, None, msg="Attribute %s is required" % attr_name)


class LDAPDuplicateObjectFound(LDAPException):
    """ Raised when looking for an unique LDAPObject and multiple one found. """

    def __init__(self, obj_type, filterstr=None, base_dn=None, scope=None):
        self.obj_type = obj_type
        self.filterstr = filterstr
        self.base_dn = base_dn
        self.scope = scope
        super().__init__("Duplicated object %s found" % obj_type.__name__)


class LDAPConnectionError(LDAPException):  # pylint: disable=too-few-public-methods
    """ Raised when fail to connect/bind to LDAP server """

    def __init__(self, uri, details=None):
        self.uri = uri
        self.details = details
        super().__init__(
            "Fail to connect on LDAP server %s (detail: %s)" % (
                uri, details if details else "No detail available"
            )
        )


class LDAPInvalidCredentials(LDAPConnectionError):
    """ Raised when trying to bind on LDAP server with invalid credentials. """

    def __init__(self, uri, bind_dn):
        self.bind_dn = bind_dn
        super().__init__(uri, "Invalid credential for %s" % bind_dn)


class LDAPDuplicatedRelationName(LDAPException):
    """ Raised when trying to add a relation on with a name already affected to another relation """

    def __init__(self, obj_type, rel_name, other_rel):
        self.obj_type = obj_type
        self.rel_name = rel_name
        self.other_rel = other_rel
        super().__init__("The relation '%s' already exist (%s) on object %s" % (rel_name, other_rel, obj_type.__name__))


class LDAPRelationAttributeConflictName(LDAPException):
    """ Raised when trying to add a relation on with a name already affected to an attribute """

    def __init__(self, obj_type, rel_name, attr):
        self.obj_type = obj_type
        self.rel_name = rel_name
        self.attr = attr
        super().__init__("The relation name '%s' is already used by an attribute (%s) on object %s" % (rel_name, attr, obj_type.__name__))


class LDAPRelationLinkAttributeValueMissing(LDAPException):
    """ Raised when trying to get link value of an object and no value is retreived """

    def __init__(self, rel, obj):
        self.rel = rel
        self.obj = obj
        super().__init__("No link attribute value found for the relation %s and the object %s" % (rel, obj))


class LDAPRelatedObjectModifed(LDAPException):
    """ Raised when trying to set related objects throught a relation and a related object is modified. """

    def __init__(self, rel, obj):
        self.rel = rel
        self.obj = obj
        super().__init__("The related object %s of the relation %s is modified. Please save changes before updating the relation" % (obj, rel))


class LDAPInvalidFilterMatchRule(LDAPException):
    """ Raised when invalid match rule is passed to compose a LDAP filter string. """

    def __init__(self, match):
        self.match = match
        super().__init__("Invalid LDAP filter string match rule '%s'" % match)


class LDAPInvalidFilterLogicalOperator(LDAPException):
    """ Raised when invalid logical operator is passed to combine LDAP filter strings. """

    def __init__(self, log_op):
        self.log_op = log_op
        super().__init__("Invalid LDAP filter string logical operator '%s'" % log_op)


class LDAPSearchLimitExceeded(LDAPException):  # pylint: disable=too-few-public-methods
    """ Raised when the search limit is exeeded """

    def __init__(self, limit):
        self.limit = limit
        super().__init__("More than %s matching object(s) found on LDAP server" % limit)


class LDAPSearchHardLimitExceeded(SIZELIMIT_EXCEEDED, LDAPSearchLimitExceeded):  # pylint: disable=too-few-public-methods
    """ Raised when the search hard limit is exeeded """

    def __init__(self, limit):
        self.limit = limit
        super().__init__("More than %s result found on LDAP server" % limit)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP filter strings helpers """

import logging

from ldap.filter import escape_filter_chars

from eesyldap.exceptions import LDAPInvalidFilterLogicalOperator
from eesyldap.exceptions import LDAPInvalidFilterMatchRule

log = logging.getLogger(__name__)


def compose(attr_name, match, raw_value=None, escape=True, escape_mode=0):
    """
    Compose an LDAP filter string

    :param  attr_name:      The LDAP attribute name to filter on
    :param  match:          The match rule. Could be one of the following:
                            - "equals" or "=" or "=="
                            - "begins"
                            - "ends"
                            - "contains"
                            - "greater" or ">"
                            - "less" or "<"
                            - "greater_or_equal" or ">="
                            - "less_or_equal" or "<="
                            - "approx" or "~="
                            - "any" or "present"
    :param  raw_value:      The raw value of the LDAP attribute (optional, not need for "any" of "present" operators)
    :param  escape:         Escape or not the provided value (optional, default: True)
    :param  escape_mode:    Escape mode (if enabled):
                            - 0 (default) : only special chars mentioned in RFC 4515 are escaped
                            - 1 : all NON-ASCII chars are escaped
                            - 2 : all chars are escaped.
    """

    # Detect negation
    negate_filter = match.startswith('not_')
    if negate_filter:
        match = match[4:]

    assert raw_value is not None or match in ('present', 'any'), "The attribute value is not need only for 'present' or 'any' match rule"
    assert isinstance(escape_mode, int) and (escape_mode >= 0 or escape_mode <= 2)
    value = escape_filter_chars(raw_value, escape_mode=escape_mode) if raw_value is not None and escape else raw_value

    if match in ("equals", "=", "=="):
        filterstr = "({0}={1})".format(attr_name, value)
    elif match == "begins":
        filterstr = "({0}={1}*)".format(attr_name, value)
    elif match == "ends":
        filterstr = "({0}=*{1})".format(attr_name, value)
    elif match == "contains":
        filterstr = "({0}=*{1}*)".format(attr_name, value)
    elif match in ("greater", ">"):
        filterstr = "({0}>{1})".format(attr_name, value)
    elif match in ("less", "<"):
        filterstr = "({0}<{1})".format(attr_name, value)
    elif match in ("greater_or_equal", ">="):
        filterstr = "({0}>={1})".format(attr_name, value)
    elif match in ("less_or_equal", "<="):
        filterstr = "({0}<={1})".format(attr_name, value)
    elif match in ("approx", "~="):
        filterstr = "({0}~={1})".format(attr_name, value)
    elif match in ("any", "present"):
        filterstr = "({0}=*)".format(attr_name)
    else:
        raise LDAPInvalidFilterMatchRule(match)

    if negate_filter:
        return combine('not', filterstr)

    return filterstr


def combine(log_op, *filterstrs):
    """
    Combine LDAP filter strings

    :param  log_op:     The logical operator of the combination. Could be one of the following:
                        - "not" or "!"
                        - "and" or "&"
                        - "or" or '|'
    :param  filterstrs: The LDAP filter strings to combine
    """
    if log_op in ('not', '!'):
        assert len(filterstrs) == 1, "Only one filter string could be combined with NOT logical operator"
        return "(!%s)" % filterstrs[0]

    # Convert 'or' / 'and' logical operators
    log_op = '&' if log_op == 'and' else log_op
    log_op = '|' if log_op == 'or' else log_op
    if log_op in ('&', '|'):
        assert len(filterstrs) >= 2, "At least two filter strings could be combined with %s logical operator" % ('AND' if log_op == '&' else 'OR')
        return "(%s%s)" % (log_op, ''.join(filterstrs))

    raise LDAPInvalidFilterLogicalOperator(log_op)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Exceptions raised during test """

from eesyldap.exceptions import LDAPInvalidAttributeValue


class CantReuseSamePassword(LDAPInvalidAttributeValue):
    """ Raised when the same password as current one is affected to a user """

    def __init__(self, attr_name, attr_value):
        super().__init__(attr_name, attr_value, msg="You can't reuse the same password")


class CantReuseOldPassword(LDAPInvalidAttributeValue):
    """ Raised when the previously used password is affected to a user """

    def __init__(self, attr_name, attr_value):
        super().__init__(attr_name, attr_value, msg="You can't reuse an old password")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Attributes """

import copy
import datetime
import json
import logging
import re

import dateutil.parser
import dateutil.tz
from passlib.hash import ldap_md5
from passlib.hash import ldap_md5_crypt
from passlib.hash import ldap_sha1
from passlib.hash import ldap_salted_sha1
from passlib.hash import ldap_sha256_crypt
from passlib.hash import ldap_sha512_crypt
import pytz

from eesyldap.exceptions import LDAPRequiredAttribute
from eesyldap.object import LDAPObject
from eesyldap.unicity import BaseUnicityCheck
from eesyldap.filters import combine as combine_ldap_filter
from eesyldap.filters import compose as compose_ldap_filter

log = logging.getLogger(__name__)
ldap_date_format = '%Y%m%d%H%M%SZ'


class Base:
    """
    LDAP attribute

    This class abstract a LDAP entry's attribute as a Python object
    """

    # Set to True if this attribute must be change in LDAP by using replace operation
    force_replace = False

    # Set to True if this attribute could only be write in LDAP (and never read)
    write_only = False

    def __init__(self, ldap_name, multiple=False, required=None, unicity=None, default_ldap=None, default_python=None,
                 force_replace=None, write_only=None):
        """
        Create a new LDAP attribute

        :param  ldap_name:          The LDAP attribute name
        :param  multiple:           Is it a multiple values attribute? Default: False
        :param  required:           Is it a required attribute? Default: False
        :param  unicity:            The value of this attribute must be unique (optional)
        :param  default_ldap:       The default value to set on LDAP server if this attribute is not set
        :param  default_python:     The default value to set on Python object if this attribute is not set on LDAP server
        :param  force_replace:      Set to True if this attribute must be change in LDAP by using replace operation
        :param  write_only:         Set to True if this attribute could only be write in LDAP (and never read)
        """
        self.ldap_name = ldap_name
        self.multiple = multiple
        self.required = required
        assert unicity is None or isinstance(unicity, BaseUnicityCheck)
        self.unicity = unicity
        if default_ldap is not None:
            if multiple:
                assert isinstance(default_ldap, list), "Default LDAP value of attribute with multiple values must be a list of string"
                for value in default_ldap:
                    assert isinstance(value, str), "Default LDAP value of attribute with multiple values must be a list of string"
            else:
                assert isinstance(default_ldap), "Default LDAP value must be a string"
        self.default_ldap = default_ldap
        self.default_python = [] if default_python is None and multiple else default_python

        if force_replace:
            self.force_replace = True

        if write_only:
            self.write_only = True

    def _to_python(self, value, sortable=None):
        """
        Convert attribute value from LDAP to Python

        :param  value:      The LDAP attribute value to convert
        :param  sortable:   If True, make sure that returned value is sortable with others
        """
        raise NotImplementedError("%s.%s" % (self.__repr__(), '_to_python'))

    def _to_ldap(self, value):
        """
        Convert attribute value from Python to LDAP

        :param  value:  The python attribute value to convert
        """
        raise NotImplementedError("%s.%s" % (self.__repr__(), '_to_ldap'))

    def get_value(self, obj=None, ldap_values=None, sortable=None):
        """
        Get python value from LDAP attribute

        :param  obj:            The LDAP object
        :param  ldap_values:    The LDAP attribute values
        :param  sortable:       If True, make sure that returned value is sortable with others
        """
        if not ldap_values and obj:
            ldap_values = obj.entry.get_values(self.ldap_name, default=self.default_ldap)

        if not ldap_values or self.write_only:
            if self.default_python is None and sortable:
                return [] if self.multiple else self._to_python(None, sortable=True)
            return self.default_python

        if self.multiple:
            return [self._to_python(value, sortable=sortable) for value in ldap_values]
        return self._to_python(ldap_values[0], sortable=sortable)

    def set_value(self, obj, value, force_replace=False):
        """
        Set LDAP value from python

        :param  obj:            The LDAP object
        :param  value:          The new attribute python value
        :param  force_replace:  Force using LDAP replace operation on change
        """
        assert (isinstance(value, list) or value is None) or not self.multiple, "%s : Multiple attribute value must be a list" % self.ldap_name
        if value is None and self.required:
            raise LDAPRequiredAttribute(self.ldap_name)
        if not self.write_only and value == self.get_value(obj):
            log.debug('%s : set value is same as current one', self.ldap_name)
        else:
            log.debug('%s : set new values to %s', self.ldap_name, value)
            if value is None:
                if self.default_ldap is not None:
                    new_values = self.default_ldap if self.multiple and (isinstance(self.default_ldap, list) or self.default_ldap is None) else [self.default_ldap]
                else:
                    new_values = None
            elif self.multiple:
                new_values = [self._to_ldap(v) for v in value]
            else:
                new_values = [self._to_ldap(value)]
            log.debug('%s : set raw new values to %s', self.ldap_name, new_values)
            if self.is_unique() and new_values is not None:
                original_values = obj.entry.get_original_values(self.ldap_name, default=[])
                for new_value in new_values:
                    if new_value in original_values:
                        #  Don't check original values of the entry (that will be only restored)
                        continue
                    if not self.unicity.is_unique(obj.client, obj, self, new_value):
                        log.warning('%s : Value "%s" for attribute %s of object %s is not unique.', self.unicity, new_value, self, obj)
                        return False
            return obj.entry.set_values(self.ldap_name, new_values, force_replace=force_replace or self.force_replace or self.write_only)
        return True

    def __repr__(self):
        """ Compute and return the “official” string representation of the attribute """
        return "<Attribute %s %s>" % (type(self).__name__, self.ldap_name)

    def __eq__(self, other):
        """ Check current attribute is equal to another """
        if not isinstance(other, self.__class__):
            return False
        return self.ldap_name == other.ldap_name

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.ldap_name)

    def is_present(self, obj):
        """ Check if attribute is present/defined """
        return obj.entry.is_present(self.ldap_name)

    def is_unique(self):
        """ Check if attribute is unique """
        return isinstance(self.unicity, BaseUnicityCheck)

    def is_multiple(self):
        """ Check if attribute is multiple """
        return bool(self.multiple)

    def get_filter(self, value, match=None):
        """
        Get/compose LDAP filter on specific value for this attribute

        :param  value:  The python value of the attribute to search on.
                        Specific values:
                        - None : attribute not present
                        - '*' : attribute present
                        - list of string : list of alternative possible values. The filter will be
                        composed with OR logical operator.
        :param  match:  Match rule as accept by eesyldap.filters.compose (optional)
        """
        # If match rule is provided, compose filter with it via eesyldap.filters.compose
        if match is not None:
            ldap_value = self._to_ldap(value).decode('UTF8') if value is not None else value
            return compose_ldap_filter(self.ldap_name, match, ldap_value)
        # Filter on None value = attribute not present
        if value is None:
            return compose_ldap_filter(self.ldap_name, 'not_present')
        # Handle special '*' value = attribute present
        if value == '*':
            return compose_ldap_filter(self.ldap_name, 'present')
        # Handle multiple values = alternative values combine with OR logical operator
        if isinstance(value, list):
            if len(value) > 1:
                filterstrs = [self.get_filter(v) for v in value]
                return combine_ldap_filter('or', *filterstrs)
            return self.get_filter(value[0])

        # Convert value to LDAP
        ldap_value = self._to_ldap(value).decode('UTF8')

        # Detect and remove heading/trailing '*' to not escape it
        star_prefix = False
        star_suffix = False
        if ldap_value.startswith('*'):
            ldap_value = ldap_value[1:]
            star_prefix = True
        if ldap_value.endswith('*'):
            ldap_value = ldap_value[:-1]
            star_suffix = True

        # Define matching rule
        if star_prefix and star_suffix:
            match = 'contains'
        elif star_prefix:
            match = 'begins'
        elif star_suffix:
            match = 'ends'
        else:
            match = 'equals'

        return compose_ldap_filter(self.ldap_name, match, ldap_value)


class Text(Base):
    """
    LDAP text attribute

    On python side, the value of this attribute is manipulated as a built-in string object.
    """

    def _to_python(self, value, sortable=None):
        """ Convert attribute value from LDAP to Python """
        if sortable and value is None:
            return ''
        return value.decode('utf-8') if value is not None else None

    def _to_ldap(self, value):
        """ Convert attribute value from Python to LDAP """
        assert isinstance(value, str), "Text value must be a string (and not a %s)" % type(value)
        return value.encode('utf-8')


class Password(Text):
    """
    LDAP password attribute

    This type automatically handle to hash the password. It also expose some helper methods to
    manage LDAP hashed password values.

    On python side, the value of this attribute is manipulated as a built-in string object.
    """

    _hash_types = ['ssha', 'sha', 'md5', 'md5_crypt', 'ssha256', 'ssha512', 'clear']

    def __init__(self, name, hash_type=None, **kwargs):
        """
        Create a new LDAP attribute

        :param  name:               The attribute name
        :param  hash_type:          Password hash type. Default: SSHA
        """
        if hash_type is None:
            self.hash_type = 'ssha'
        else:
            assert hash_type in self._hash_types, 'Hash type incorrect. Must one of the following types : %s' % ', '.join(self._hash_types)
            self.hash_type = hash_type
        super().__init__(name, **kwargs)

    def verify(self, clear_password, hashed_passwords=None, obj=None):  # pylint: disable=too-many-return-statements
        """
        Verify the correspondence between an clear password and a hashed one

        :param  clear_password:     The clear password to check against hashed ones.
        :param  hashed_passwords:   The list of hashed passwords (Default: the attribute values if obj parameter is provided)
        :param  obj:                The LDAP object (optional and not used if hashed_passwords parameter is provided)
        """
        if hashed_passwords is None:
            assert isinstance(obj, LDAPObject), "If you don't specify hashed_passwords parameter, you must specify obj parameter"
            hashed_passwords = self.get_value(obj)
            if not hashed_passwords:
                log.debug('Password attribute %s : verify failed because the value of the attribute is empty', self.ldap_name)
                return False
            # log.debug('Password values retreived from object entry = %s', hashed_passwords)

        if not isinstance(hashed_passwords, list):
            hashed_passwords = [hashed_passwords]

        for hashed_password in hashed_passwords:
            hash_type = re.match('^{([^}]+)}', hashed_password)
            if not hash_type:
                # log.debug('Check clear password "%s" against another clear password "%s"', clear_password, hashed_password)
                return clear_password == hashed_password
            hash_type = hash_type.group(1)
            # log.debug('Check clear password "%s" against %s hashed password "%s"', clear_password, hash_type, hashed_password)
            if hash_type == 'MD5':
                if ldap_md5.verify(clear_password, hashed_password):
                    return True
                continue
            if hash_type == 'SHA':
                if ldap_sha1.verify(clear_password, hashed_password):
                    return True
                continue
            if hash_type == 'SSHA':
                if ldap_salted_sha1.verify(clear_password, hashed_password):
                    return True
                continue
            if hashed_password.startswith('{CRYPT}$1$'):
                if ldap_md5_crypt.verify(clear_password, hashed_password):
                    return True
                continue
            if hashed_password.startswith('{CRYPT}$5$'):
                if ldap_sha256_crypt.verify(clear_password, hashed_password):
                    return True
                continue
            if hashed_password.startswith('{CRYPT}$6$'):
                if ldap_sha512_crypt.verify(clear_password, hashed_password):
                    return True
                continue

            log.warning('Unrecognized hash password type (%s)', hash_type)

        return False

    @staticmethod
    def is_hashed_password(value):
        """ Check if the value is a hashed password """
        if not isinstance(value, str):
            return False
        if re.match('^({[^}]+})', value):
            return True
        return False

    def hash(self, password, hash_type=None):
        """
        Hash a password

        :param  password:       The password to hash
        :param  hash_type:      The hash type (optional, default: as configured in the constructor)
        """
        if hash_type is None:
            hash_type = self.hash_type

        log.debug("Hash password '%s' as %s", password, hash_type)

        if hash_type == 'clear':
            return password
        if hash_type == 'md5':
            return ldap_md5.hash(password)
        if hash_type == 'md5_crypt':
            return ldap_md5_crypt.hash(password)
        if hash_type == 'sha':
            return ldap_sha1.hash(password)
        if hash_type == 'ssha':
            return ldap_salted_sha1.hash(password)
        if hash_type == 'ssha256':
            return ldap_sha256_crypt.hash(password)
        if hash_type == 'ssha512':
            return ldap_sha512_crypt.hash(password)

        raise Exception('Unknwon hash type %s' % hash_type)

    def _to_ldap(self, value):
        """ Convert attribute value from Python to LDAP """
        return super()._to_ldap(value if self.is_hashed_password(value) else self.hash(value))


class Integer(Base):
    """
    LDAP integer attribute

    On python side, the value of this attribute is manipulated as a built-in integer object.
    """

    def _to_python(self, value, sortable=None):
        """ Convert attribute value from LDAP to Python """
        if sortable and value is None:
            return int()
        return int(value) if value is not None else None

    def _to_ldap(self, value):
        """ Convert attribute value from Python to LDAP """
        assert isinstance(value, int), "Integer value must be a integer (and not a %s)" % type(value)
        return str(value).encode('UTF8')


class Boolean(Base):
    """
    LDAP boolean attribute

    This type handle standard LDAP boolean attribute that stores the TRUE or FALSE values.

    On python side, the value of this attribute is manipulated as a built-in boolean object.
    """

    def _to_python(self, value, sortable=None):
        """ Convert attribute value from LDAP to Python """
        if sortable:
            # Return a sortable string value
            if value is None:
                return ''
            return 1 if value == b'TRUE' else 0
        return value == b'TRUE' if value is not None else None

    def _to_ldap(self, value):
        """ Convert attribute value from Python to LDAP """
        return b'TRUE' if value else b'FALSE'


class Datetime(Base):
    """
    LDAP datetime attribute

    This type handle standard LDAP Generalized Time attribute.

    On python side, the value of this attribute is manipulated as a datetime.datetime object.
    """

    def __init__(self, name, python_timezone=None, default_python_timezone=None, ldap_timezone=None, default_ldap_timezone=None, naive=None, **kwargs):
        """
        Create a new LDAP attribute

        :param  name:                       The attribute name
        :param  python_timezone:            Timezone use on python side (Default: local)
        :param  default_python_timezone:    Default timezone used if naive datetime is pass from python
                                            (Default: python_timezone)
        :param  ldap_timezone:              Timezone use on LDAP side (Default: UTC)
        :param  default_ldap_timezone:      Default timezone used if naive datetime is retreived from LDAP
                                            (not supposed to happen, default: ldap_timezone)
        :param  naive:                      Use naive datetime object : do not handle timezone on both Python
                                            and LDAP sides. Datetime are stored as UTC format on LDAP and Python
                                            datetime objects are naive.

        Timezone parameters possible values :
          - 'local' : Local server configured timezone
          - any value recognized by pytz.timezone()
        """
        self.python_timezone = self._convert_timezone_parameter('python_timezone', python_timezone, dateutil.tz.tzlocal())
        self.default_python_timezone = self._convert_timezone_parameter('default_python_timezone', default_python_timezone, self.python_timezone)
        self.ldap_timezone = self._convert_timezone_parameter('ldap_timezone', ldap_timezone, pytz.utc)
        self.default_ldap_timezone = self._convert_timezone_parameter('default_ldap_timezone', default_ldap_timezone, self.ldap_timezone)
        self.naive = naive
        super().__init__(name, **kwargs)

    @staticmethod
    def _convert_timezone_parameter(name, value, default):
        """ Convert constructor timezone parameter to timezone value """
        if value is None:
            return default
        if isinstance(value, str):
            if value == 'local':
                return dateutil.tz.tzlocal()
            return pytz.timezone(value)
        if isinstance(value, datetime.tzinfo):
            return value
        raise ValueError('Parameter %s must be a string or datetime.tzinfo object (not %s)' % (name, type(value)))

    def _to_python(self, value, sortable=None):
        """ Convert attribute value from LDAP to Python """
        if value is None:
            return '' if sortable else None
        date = dateutil.parser.parse(value, dayfirst=False)
        # Check if datetime
        if not date.tzinfo and not self.naive:
            if isinstance(self.default_ldap_timezone, pytz.tzinfo.DstTzInfo):
                date = self.default_ldap_timezone.localize(date)
            else:
                date = date.replace(tzinfo=self.default_ldap_timezone)
        elif self.naive:
            date = date.replace(tzinfo=None)
        if self.python_timezone and not self.naive:
            return date.astimezone(self.python_timezone)
        if sortable:
            if not self.naive:
                # Force UTC timezone (to make dates sortable together regardless their timezone)
                date = date.astimezone(pytz.utc)
            return date.strftime('%Y%m%d%H%M%S')
        return date

    def _to_ldap(self, value):
        """ Convert attribute value from Python to LDAP """
        assert isinstance(value, datetime.datetime), "Date value must be a datetime.datetime (and not a %s)" % type(value)
        if not value.tzinfo and not self.naive:
            if isinstance(self.default_python_timezone, pytz.tzinfo.DstTzInfo):
                from_value = self.default_python_timezone.localize(value)
            else:
                from_value = value.replace(tzinfo=self.default_python_timezone)
        elif self.naive:
            from_value = value.replace(tzinfo=None)
        else:
            from_value = copy.deepcopy(value)
        to_value = from_value.astimezone(self.ldap_timezone) if not self.naive else from_value
        datestring = to_value.strftime('%Y%m%d%H%M%S%z')
        if self.naive:
            datestring += 'Z'
        elif datestring.endswith('+0000'):
            datestring = datestring.replace('+0000', 'Z')
        return datestring.encode('utf-8')


class Date(Datetime):
    """
    LDAP date attribute

    This type handle standard LDAP Generalized Time attribute that store a date at midnight in UTC timezone.
    It derive from Datetime attribute type and support all of it parameters.

    On python side, the value of this attribute is manipulated as a datetime.date object.
    """

    def __init__(self, name, **kwargs):
        """
        Create a new LDAP attribute

        :param  name:                       The attribute name
        """
        kwargs['naive'] = True
        for kwarg in ['python_timezone', 'default_python_timezone', 'ldap_timezone', 'default_ldap_timezone']:
            assert kwarg not in kwargs, 'Date attribute is timezone naive on both LDAP and Python sides'
        super().__init__(name, **kwargs)

    def _to_python(self, value, sortable=None):
        """ Convert attribute value from LDAP to Python """
        if value is None:
            return '' if sortable else None
        if sortable:
            return super()._to_python(value, sortable=True)
        return super()._to_python(value).date()

    def _to_ldap(self, value):
        """ Convert attribute value from Python to LDAP """
        assert isinstance(value, datetime.date), "Date value must be a datetime.date (and not a %s)" % type(value)
        return super()._to_ldap(datetime.datetime.combine(value, datetime.datetime.min.time()))


class ShadowLastChange(Integer):
    """
    LDAP shadowLastChange attribute

    This type handle standard LDAP shadowLastChange attribute that store the date of the
    last password changed as an integer corresponding to number of days since 1970/01/01.

    On python side, the value of this attribute is manipulated as a datetime.date object.
    """

    def _to_python(self, value, sortable=None):
        value = super()._to_python(value)
        if value is None:
            return '' if sortable else None
        date = datetime.date(1970, 1, 1) + datetime.timedelta(days=value)
        if sortable:
            return date.strftime('%Y%m%d')
        return date

    def _to_ldap(self, value):
        assert isinstance(value, datetime.date), "Date value must be a datetime.date (and not a %s)" % type(value)
        return super()._to_ldap((value - datetime.date(1970, 1, 1)).days)


class JSON(Text):
    """
    LDAP JSON composited attribute

    On python side, the value of this attribute is manipulated as an object (or a list of objects
    if this attribute is multiple) as serialized in JSON.
    """

    def _to_python(self, value, sortable=None):
        if value is None and sortable:
            return ''
        value = super()._to_python(value, sortable=sortable)
        if value is None:
            return None
        return json.loads(value)

    def _to_ldap(self, value):
        value = json.dumps(value)
        return super()._to_ldap(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Object """

from copy import deepcopy
import logging

from ldap.dn import escape_dn_chars
from ldap.dn import explode_dn

log = logging.getLogger(__name__)


class LDAPEntry:
    """
    LDAP entry

    This class abstract a LDAP entry as a Python object
    """

    # Client reference
    _client = None

    # LDAP object reference
    _obj = None

    # LDAP object DN
    _dn = None

    # LDAP attributes values
    _attrs_values = None

    # LDAP attributes new values (on changed)
    _new_attrs_values = None

    # List LDAP attributes for which we must use the replace operation on change
    _force_replace_attrs = None

    def __init__(self, client, obj, dn=None, attrs_values=None):
        self._client = client
        self._obj = obj
        self._dn = dn
        self._attrs_values = dict(
            (attr_name.lower(), attr_values)
            for attr_name, attr_values in (attrs_values or {}).items()
        )
        self._new_attrs_values = {}
        self._force_replace_attrs = []

    def copy(self):
        """ Create a copy of the current LDAPEntry object """
        new_entry = self.__class__(self._client, self._obj, dn=deepcopy(self._dn), attrs_values=deepcopy(self._attrs_values))
        new_entry._new_attrs_values = deepcopy(self._new_attrs_values)  # pylint: disable=protected-access
        new_entry._force_replace_attrs = deepcopy(self._force_replace_attrs)  # pylint: disable=protected-access
        return new_entry

    def get_values(self, attr, default=None):
        """ Get an attributes's values """
        attr = attr.lower()
        # log.debug('%s.get_values(%s) : cur = %s / new = %s', self, attr, self._attrs_values.get(attr), self._new_attrs_values.get(attr))
        return self._new_attrs_values.get(attr, self.get_original_values(attr, default=default))

    def get_original_values(self, attr, default=None):
        """ Get an attributes's orignal values """
        attr = attr.lower()
        return self._attrs_values.get(attr, default)

    def set_values(self, attr, values, force_replace=False):
        """ Set an attributes's values """
        assert isinstance(values, list) or values is None, "Entry attributes values must a list or None (%s given for attribute %s)" % (type(values), attr)
        attr = attr.lower()
        # log.debug('%s.set_values(%s, %s) : cur = %s / new = %s', self, attr, values, self._attrs_values.get(attr), self._new_attrs_values.get(attr))
        if self._new_attrs_values.get(attr):
            if values == self._new_attrs_values[attr]:
                log.debug('%s.set_values(%s, %s) : new values same as current new one', self, attr, values)
                return True
            # Try to restore current LDAP value ?
            if values == self._attrs_values.get(attr) and not force_replace:
                log.debug('%s.set_values(%s, %s) : new values same as current LDAP one => restore it', self, attr, values)
                del self._new_attrs_values[attr]
                if attr in self._force_replace_attrs:
                    self._force_replace_attrs.remove(attr)
                return True
        elif values == self._attrs_values.get(attr) and not force_replace:
            log.debug('%s.set_values(%s, %s) : new values same as current one', self, attr, values)
            return True
        self._new_attrs_values[attr] = values
        if force_replace and attr not in self._force_replace_attrs:
            self._force_replace_attrs.append(attr)
        return True

    def is_present(self, attr):
        """ Check if an attribute is present/defined """
        attr = attr.lower()
        if attr in self._new_attrs_values:
            return self._new_attrs_values[attr] is not None
        return attr in self._attrs_values

    def is_modified(self, attr=None):
        """ Check if the entry (or one of this attribute) is modified """
        return bool(self._new_attrs_values) if attr is None else (attr in self._new_attrs_values)

    def get_dn(self, new_rdn=None):
        """
        Retreive or generate LDAP object DN

        :param  new_rdn:    The LDAP entry new RDN value (ex : uid=jdoe)
        """
        if self._dn and not new_rdn:
            return self._dn

        # If new RDN is provided, compose DN with it
        if new_rdn:
            exploded_dn = explode_dn(self._dn)
            return "%s,%s" % (new_rdn, ','.join(exploded_dn[1:]))

        # Otherwise, compose DN with first RDN attribute value
        rdn_values = self.get_values(self._obj.rdn_ldap_attr)
        if not rdn_values:
            return False
        base_dn = self._client.get_object_base_dn(self._obj.__class__)
        return "%s=%s,%s" % (self._obj.rdn_ldap_attr, escape_dn_chars(rdn_values[0].decode('utf-8')), base_dn)

    def save(self):
        """ Save changes on the LDAP entry """

        # Generate old / new attrs dict (only changed attributes)
        old_attrs = {}
        new_attrs = {}
        for attr in self._new_attrs_values:
            if self._new_attrs_values[attr] is not None:
                new_attrs[attr] = self._new_attrs_values[attr]
            if attr in self._attrs_values:
                old_attrs[attr] = self._attrs_values[attr]

        rdn_attr = self._obj.rdn_ldap_attr.lower()

        if self._dn is None:
            # New object
            dn = self.get_dn()
            new_attrs['objectClass'] = [c.encode('utf8') for c in self._obj.classes]
            log.debug('Object is new : add it with DN "%s" and the following attributes : %s', dn, new_attrs)
            if self._client.add(self._obj, dn, new_attrs):
                self._saved(new_values=new_attrs, new_dn=dn)
                return True
        elif rdn_attr in new_attrs:
            # Existing LDAP entry and need to be renamed
            new_rdn = "%s=%s" % (rdn_attr, escape_dn_chars(new_attrs[rdn_attr][0].decode('utf8')))
            log.debug('The RDN of the object %s changed : rename it with new RDN "%s"', self._obj.dn, new_rdn)
            if self._client.rename(self._obj, new_rdn):
                # Set RDN attr and DN as saved
                self._saved(
                    new_values=new_attrs[rdn_attr],
                    old_values=old_attrs[rdn_attr],
                    attr=rdn_attr,
                    new_dn=self.get_dn(new_rdn=new_rdn)
                )

                # Remove RDN attr in old / new attrs
                del old_attrs[rdn_attr]
                del new_attrs[rdn_attr]
            else:
                log.debug('Fail to rename object : do not continue')
                return False

        if old_attrs or new_attrs:
            # Existing object with changes
            log.debug('%s : save() : old = %s / new = %s', self, old_attrs, new_attrs)
            if self._client.modify(self._obj, old_attrs, new_attrs, force_replace_attrs=self._force_replace_attrs):
                self._saved(new_values=new_attrs, old_values=old_attrs)
                return True
            return False

        # Existing object without changes
        log.debug('%s : save() : no changes', self._dn)
        log.debug('%s : no changes to save', self._dn)
        return True

    def _saved(self, new_values, old_values=None, new_dn=None, attr=None):
        """
        Set entry as saved

        :param  new_values:     New attributes values (only attr value if attr is set)
        :param  new_dn:         New DN (optional)
        :param  attr:           Attribute name if saved changed is about only one attribute (optional)
        """
        if new_dn:
            log.debug('Set new LDAP entry DN to "%s"', new_dn)
            self._dn = new_dn
        if attr is not None:
            log.debug('Set changed of attribute %s as saved : %s => %s', attr, old_values or 'Unknwon', new_values)
            if new_values is None:
                del self._attrs_values[attr]
            else:
                self._attrs_values[attr] = new_values
            del self._new_attrs_values[attr]
            if attr in self._force_replace_attrs:
                self._force_replace_attrs.remove(attr)
        else:
            for attr_name, attr_values in new_values.items():
                log.debug(
                    'Set changed of attribute %s as saved : %s => %s',
                    attr_name,
                    old_values.get(attr_name, 'Undefined') if old_values else 'Unknwon',
                    attr_values or 'Undefined'
                )
                self._attrs_values[attr_name] = attr_values
                if attr_name in self._new_attrs_values:
                    del self._new_attrs_values[attr_name]
                if attr_name in self._force_replace_attrs:
                    self._force_replace_attrs.remove(attr_name)
            if old_values:
                for attr_name, attr_values in old_values.items():
                    if attr_name not in new_values:
                        log.debug('Set changed of attribute %s as deleted', attr_name)
                        del self._attrs_values[attr_name]
                    if attr_name in self._new_attrs_values:
                        del self._new_attrs_values[attr_name]
                    if attr_name in self._force_replace_attrs:
                        self._force_replace_attrs.remove(attr_name)

    def restore(self):
        """ Restore LDAP entry at initial state """
        self._new_attrs_values = {}
        return True

    def __repr__(self):
        """ Compute and return the “official” string representation of the LDAPEntry """
        return "<LDAPEntry %s>" % (self._dn or '[new]')

    def __getstate__(self):
        """ Retreive the state of the LDAP entry object for pickle """
        return dict(
            _dn=deepcopy(self._dn),
            _attrs_values=deepcopy(self._attrs_values),
            _new_attrs_values=deepcopy(self._new_attrs_values),
        )

    def __setstate__(self, state):
        """ Set the state of the LDAP entry object from pickle state """
        self.__dict__.update(state)

    def serialize(self):
        """ Retreive the serialized state of the LDAP entry object """
        return self.__getstate__()

    @classmethod
    def unserialize(cls, client, obj, state):
        """ Retreived on object from it serialized state and client and objects references """
        entry = cls(client, obj)
        entry.__setstate__(state)
        return entry



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EesyLDAP -- Simple LDAP Objects mapper
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2019 Benjamin Renard, Easter-eggs
# https://gitlab.com/brenard/eesyldap
#
# This file is part of EesyLDAP.
#
# EesyLDAP is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EesyLDAP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" LDAP Object relations """

import logging

from eesyldap.exceptions import LDAPDuplicateObjectFound
from eesyldap.exceptions import LDAPRelationLinkAttributeValueMissing
from eesyldap.exceptions import LDAPRelatedObjectModifed

log = logging.getLogger(__name__)


class RelationOnLocalLinkAttribute:
    """
    LDAP object relation using local object link attribute

    This type of relation is based on a local attribute that stores one value (or more)
    that refers specifically to another object.
    """

    def __init__(self, obj_type, link_attr, link_attr_value='dn', key_attr=None,
                 reverse_name=None, reverse_multiple=True, reverse_key_attr=None):
        """
        Relation constructor

        :param  obj_type:           The related object type
        :param  link_attr:          The local link attribute name
        :param  link_attr_value:    The name of the attribute of the related object type whose value is stored by the
                                    link attribute (default: dn)
        :param  key_attr:           The name of the key attribute (on multiple relation = if the local link attribute
                                    is multiple)
        :param  reverse_name:       The name of the reversed relation to create on related object type (optinal)
        :param  reverse_multiple:   Boolean to define if the reversed relation is multiple
        :param  reverse_key_attr:   The key attribute of the reversed relation (if multiple)
        """
        self.obj_type = obj_type
        self.link_attr = link_attr
        self.link_attr_value = link_attr_value
        assert link_attr_value == 'dn' or obj_type.exists(link_attr_value), "%s : Object type %s have no attribute '%s' : " % (self, obj_type.__name__, link_attr_value)
        if key_attr is not None:
            assert key_attr == 'dn' or obj_type.exists(key_attr), "%s : Object type %s have no attribute '%s' : can't use it as key attribute" % (self, obj_type.__class__, key_attr)
            self.key_attr = key_attr
        else:
            self.key_attr = obj_type.get_rdn_attr() or 'dn'
        self.reverse_name = reverse_name
        self.reverse_multiple = reverse_multiple
        self.reverse_key_attr = reverse_key_attr

    def __repr__(self):
        """ Compute and return the “official” string representation of the relation object """
        return "<%s on %s=%s.%s>" % (self.__class__.__name__, self.link_attr, self.obj_type.__name__, self.link_attr_value)

    def add_reverse_relation(self, obj_type):
        """
        Add reverse relation if name is defined

        :param  obj_type:       The local object type
        """
        if self.reverse_name:
            reverse_relation = RelationOnRemoteLinkAttribute(
                obj_type=obj_type,
                link_attr=self.link_attr,
                link_attr_value=self.link_attr_value,
                multiple=self.reverse_multiple,
                key_attr=self.reverse_key_attr
            )
            log.debug('%s : Add reverse relation "%s" on %s object type : %s', self, self.reverse_name, self.obj_type.__name__, reverse_relation)
            self.obj_type.add_relation(self.reverse_name, reverse_relation)

    def _get_link_value(self, r_obj):
        """ Retreive the link value of a specific object and raise LDAPRelationLinkAttributeValueMissing
        if no link value is available """
        # Retreive and check related objetc link value
        link_value = getattr(r_obj, self.link_attr_value)
        if not link_value:
            raise LDAPRelationLinkAttributeValueMissing(self, r_obj)
        return link_value

    def get_related_objects(self, obj):
        """ Retreive related object(s) of a specific object """

        # Check that link attr exists
        assert obj.exists(self.link_attr), "%s : Object type %s have no attribute '%s' : " % (self, obj.__name__, self.link_attr)

        multiple = obj.is_multiple(self.link_attr)

        # Check link_attr_value
        link_values = getattr(obj, self.link_attr)
        if not link_values:
            log.debug('No link value retreive from "%s"', self.link_attr)
            return None if not multiple else dict()

        # If multiple, iter on link_values
        if multiple:
            result = dict()
            for link_value in link_values:
                identifiers = {self.link_attr_value: link_value}
                r_obj = obj.client.get(self.obj_type, **identifiers)
                if r_obj:
                    result[getattr(r_obj, self.key_attr) if self.key_attr else link_value] = r_obj
                else:
                    log.warning("%s : No %s related object found for %s=%s", self, self.obj_type.__name__, self.link_attr_value, link_value)
            return result

        # Otherwise :
        identifiers = {self.link_attr_value: link_values}
        return obj.client.get(self.obj_type, **identifiers)

    def set_related_objects(self, obj, related_objs):
        """
        Set related object of a specific object

        :param  obj:            The local object
        :param  related_objs:   The related objects : if the relation is multiple, it must be a dict as retreived by
                                the get_related_objects() method. If the relation is not multiple, it's must be an object
                                of the type of related object. In any case, it could be None (or an empty dict) to remove
                                all relations with current related objects.
        """
        # Check that link attr exists
        assert obj.exists(self.link_attr), "%s : Object type %s have no attribute '%s' : " % (self, obj.__name__, self.link_attr)

        # Verify that object is not modified
        if obj.is_modified():
            raise LDAPRelatedObjectModifed(self, obj)

        # If not related_objs, remove link attr
        if not related_objs:
            link_value = None
        elif obj.is_multiple(self.link_attr):
            # On multiple case
            assert isinstance(related_objs, dict), "%s : This relation is multiple : you must provide a dict of %s objects with %s attribute as key" % (self, self.obj_type.__name__, self.key_attr or 'default')
            link_value = []
            for r_obj_id, r_obj in related_objs.items():    # pylint: disable=unused-variable
                assert isinstance(r_obj, self.obj_type), "%s : You must provide a list of %s objects (%s provided)" % (self, self.obj_type.__name__, type(r_obj).__name__)

                # Check related object is not modified
                if r_obj.is_modified():
                    raise LDAPRelatedObjectModifed(self, r_obj)

                # Retreive and check related objetc link value
                r_obj_value = self._get_link_value(r_obj)

                link_value.append(r_obj_value)
        else:
            # On non-multiple case
            assert isinstance(related_objs, self.obj_type), "%s : You must provide a %s object (%s provided)" % (self, self.obj_type, type(related_objs).__name__)

            # Check related object is not modified
            if related_objs.is_modified():
                raise LDAPRelatedObjectModifed(self, related_objs)

            # Retreive and check related objetc link value
            link_value = self._get_link_value(related_objs)

        # Set link attribute on the local object
        setattr(obj, self.link_attr, link_value)
        return obj.save()

    def add_related_object(self, obj, r_obj, link_value=None):
        """
        Add a related object

        :param  obj:        The local object
        :param  r_obj:      The related object
        :param  link_value: The link value that refers to the related object (optional)
        """
        # Check that link attr exists
        assert obj.exists(self.link_attr), "%s : Object type %s have no attribute '%s' : " % (self, obj.__name__, self.link_attr)

        # Check object is not modified
        if obj.is_modified():
            raise LDAPRelatedObjectModifed(self, obj)

        # Check related object is not modified
        if r_obj.is_modified():
            raise LDAPRelatedObjectModifed(self, r_obj)

        # Retreive and check related objetc link value
        r_obj_value = self._get_link_value(r_obj)

        if obj.is_multiple(self.link_attr):
            link_value = getattr(obj, self.link_attr)
            if r_obj_value in link_value:
                # This relation already exists between obj and r_obj
                return True
            link_value.append(r_obj_value)
        else:
            link_value = r_obj_value

        # Set link attribute on the local object
        setattr(obj, self.link_attr, link_value)
        return obj.save()

    def remove_related_object(self, obj, r_obj, link_value=None):
        """
        Remove a related object

        :param  obj:        The local object
        :param  r_obj:      The related object
        :param  link_value: The link value that refers to the related object (optional)
        """
        # Check that link attr exists
        assert obj.exists(self.link_attr), "%s : Object type %s have no attribute '%s' : " % (self, obj.__name__, self.link_attr)

        # Check object is not modified
        if obj.is_modified():
            raise LDAPRelatedObjectModifed(self, obj)

        # Check related object is not modified
        if r_obj.is_modified():
            raise LDAPRelatedObjectModifed(self, r_obj)

        # Retreive and check related objetc link value
        r_obj_value = self._get_link_value(r_obj)

        if obj.is_multiple(self.link_attr):
            link_value = getattr(obj, self.link_attr)
            if r_obj_value not in link_value:
                # This relation does not exists between obj and r_obj
                return True
            link_value.remove(r_obj_value)
        else:
            link_value = None

        # Set link attribute on the local object
        setattr(obj, self.link_attr, link_value)
        return obj.save()

    def handle_object_changes(self, old_obj, new_obj):  # pylint: disable=unused-argument,no-self-use
        """
        Handle changes on object

        :param  old_obj:    The object in its state before modification
        :param  new_obj:    The object in its state after modification
        """
        # Link values are store on local object attributes, no changes need on related objects
        return True

    def handle_object_removal(self, obj):   # pylint: disable=unused-argument,no-self-use
        """
        Handle removal of object

        :param  obj:    The removed object
        """
        # Link values are store on local object attributes, no changes need on related objects
        return True


class RelationOnRemoteLinkAttribute:
    """
    LDAP object relation using remote object link attribute

    This type of relation is based on an attribute of the related objects that stores one value that
    refers specifically to the local object.
    """

    def __init__(self, obj_type, link_attr, link_attr_value='dn', multiple=True, key_attr=None,
                 reverse_name=None, reverse_key_attr=None):
        """
        Relation constructor

        :param  obj_type:           The related object type
        :param  link_attr:          The link attribute name on related object
        :param  link_attr_value:    The name of the local attribute whose value is stored by the link attribute
                                    of the related object type (default: dn)
        :param  multiple:           Boolean to define if the relation is multiple
        :param  key_attr:           The name of the key attribute (on multiple relation)
        :param  reverse_name:       The name of the reversed relation to create on related object type (optinal)
        :param  reverse_key_attr:   The key attribute of the reversed relation (if multiple)
        """
        self.obj_type = obj_type
        self.link_attr = link_attr
        self.link_attr_value = link_attr_value
        assert obj_type.exists(link_attr), "%s : Object type %s have no attribute '%s' : " % (self, obj_type.__name__, link_attr)
        self.multiple = multiple
        assert key_attr is None or not multiple, "%s : The key_attr could only be used with multiple relation" % self
        if key_attr is not None:
            assert key_attr == 'dn' or obj_type.exists(key_attr), "%s : Object type %s have no attribute '%s' : can't use it as key attribute" % (self, obj_type.__class__, key_attr)
            self.key_attr = key_attr
        else:
            self.key_attr = obj_type.get_rdn_attr() or 'dn'
        self.reverse_name = reverse_name
        self.reverse_key_attr = reverse_key_attr

    def add_reverse_relation(self, obj_type):
        """
        Add reverse relation if name is defined

        :param  obj_type:       The local object type
        """
        if self.reverse_name:
            reverse_relation = RelationOnLocalLinkAttribute(
                obj_type=obj_type,
                link_attr=self.link_attr,
                link_attr_value=self.link_attr_value,
                key_attr=self.reverse_key_attr
            )
            log.debug('%s : Add reverse relation "%s" on %s object type : %s', self, self.reverse_name, self.obj_type.__name__, reverse_relation)
            self.obj_type.add_relation(self.reverse_name, reverse_relation)

    def __repr__(self):
        """ Compute and return the “official” string representation of the relation object """
        return "<%s on %s.%s=%s>" % (self.__class__.__name__, self.obj_type.__name__, self.link_attr, self.link_attr_value)

    def _get_link_value(self, obj):
        """ Retreive the link value of a specific object and raise LDAPRelationLinkAttributeValueMissing
        if no link value is available """
        # Check link_attr_value
        assert self.link_attr_value == 'dn' or obj.exists(self.link_attr_value), "%s : Invalid link_attr_value '%s' provided in constructor" % (self, self.link_attr_value)
        link_value = getattr(obj, self.link_attr_value)
        if not link_value:
            raise LDAPRelationLinkAttributeValueMissing(self, obj)
        return link_value

    def get_related_objects(self, obj):
        """ Retreive related object(s) of a specific object """
        # Check link_attr_value
        assert self.link_attr_value == 'dn' or obj.exists(self.link_attr_value), "Invalid link_attr_value '%s' provided in constructor" % self.link_attr_value
        try:
            link_value = self._get_link_value(obj)
        except LDAPRelationLinkAttributeValueMissing:
            log.debug('No link value retreive from "%s"', self.link_attr_value)
            return None if not self.multiple else dict()

        # Composed search filters of related objects
        filters = {self.link_attr: link_value}

        # Run search
        result = obj.client.search(self.obj_type, key_attr=self.key_attr, **filters)

        # If multiple, return result or an empty dict()
        if self.multiple:
            return result

        # If not multiple, check result count
        if len(result) > 1:
            raise LDAPDuplicateObjectFound(self.obj_type)

        return result.first() if result else None

    def set_related_objects(self, obj, related_objs):
        """
        Set related object of a specific object

        :param  obj:            The local object
        :param  related_objs:   The related objects : if the relation is multiple, it must be a dict as retreived by
                                the get_related_objects() method. If the relation is not multiple, it's must be an object
                                of the type of related object. In any case, it could be None (or an empty dict) to remove
                                all relations with current related objects.
        """
        # Check link_attr_value
        assert self.link_attr_value == 'dn' or obj.exists(self.link_attr_value), "%s : Invalid link_attr_value '%s' provided in constructor" % (self, self.link_attr_value)
        link_value = self._get_link_value(obj)

        # Check object is not modified
        if obj.is_modified():
            raise LDAPRelatedObjectModifed(self, obj)

        # Retreive current list of related objects
        current_related_objects = self.get_related_objects(obj)

        # Check if no changes
        if related_objs == current_related_objects:
            log.debug('%s : no changes in related objects with %s', self, obj)
            return True

        # Init error flag
        error = False

        if self.multiple:
            # On multiple case
            assert isinstance(related_objs, dict), "%s : This relation is multiple : you must provide a dict of %s objects with %s attribute as key" % (self, self.obj_type.__name__, self.key_attr or 'defaut')

            # Iter on new related objects to add relation
            for r_obj_id, r_obj in related_objs.items():
                assert isinstance(r_obj, self.obj_type), "%s : You must provide a list of %s objects (%s provided)" % (self, self.obj_type.__name__, type(r_obj).__name__)
                log.debug('%s : Add relation between %s and %s', self, obj, r_obj)
                if not self.add_related_object(obj, r_obj, link_value=link_value):
                    error = True

            # Look for removed relation
            for r_obj_id, r_obj in current_related_objects.items():
                if r_obj_id not in related_objs:
                    log.debug('%s : Remove relation between %s and %s', self, obj, r_obj)
                    if not self.remove_related_object(obj, r_obj, link_value=link_value):
                        error = True
        else:
            # On non multiple case
            assert isinstance(related_objs, self.obj_type), "%s : You must provide a %s objects (%s provided)" % (self, self.obj_type.__name__, type(related_objs).__name__)

            if not self.add_related_object(obj, related_objs, link_value=link_value):
                error = True

            if not self.remove_related_object(obj, related_objs, link_value=link_value):
                error = True

        return not error

    def add_related_object(self, obj, r_obj, link_value=None):
        """
        Add a related object

        :param  obj:        The local object
        :param  r_obj:      The related object
        :param  link_value: The link value that refers to the local object (optional)
        """
        # Check related object is not modified
        if r_obj.is_modified():
            raise LDAPRelatedObjectModifed(self, r_obj)

        # Retreive link value
        link_value = self._get_link_value(obj) if link_value is None else link_value

        changed = False
        cur_value = getattr(r_obj, self.link_attr)
        if r_obj.is_multiple(self.link_attr):
            new_value = cur_value or []
            if link_value not in new_value:
                new_value.append(link_value)
                changed = True
        elif cur_value != link_value:
            new_value = link_value
            changed = True
        if changed:
            # Update remote object link attribute and save it
            log.debug('%s : %s -> %s', self, cur_value, new_value)
            setattr(r_obj, self.link_attr, new_value)
            if not r_obj.save():
                log.error("%s : An error occured updating attr %s of the object %s with values %s to link it with object %s", self, self.link_attr, r_obj, new_value, obj)
                return False
            log.debug('%s : Relation between %s and %s added', self, r_obj, obj)
        else:
            log.debug('%s : Object %s still in relation with %s', self, r_obj, obj)
        return True

    def remove_related_object(self, obj, r_obj, link_value=None):
        """
        Remove a related object

        :param  obj:        The local object
        :param  r_obj:      The related object
        :param  link_value: The link value that refers to the local object (optional)
        """
        # Retreive link value
        link_value = self._get_link_value(obj) if link_value is None else link_value

        if r_obj.is_multiple(self.link_attr):
            new_value = getattr(r_obj, self.link_attr) or []
            if link_value in new_value:
                new_value.remove(link_value)
        else:
            new_value = None
        # Update remote object link attribute and save it
        setattr(r_obj, self.link_attr, new_value)
        if not r_obj.save():
            log.error("%s : An error occured updating attr %s of the object %s with values %s to unlink it with object %s", self, self.link_attr, r_obj, new_value, obj)
            return False
        log.debug('%s : Relation between %s and %s removed', self, r_obj, obj)
        return True

    def rename_obj_on_related_object(self, obj, r_obj, old_link_value, new_link_value):
        """
        Rename object on related object link attribute

        :param  obj:            The local object
        :param  r_obj:          The related object
        :param  old_link_value: The old link value that refers to the local object
        :param  new_link_value: The new link value that refers to the local object
        """
        if r_obj.is_multiple(self.link_attr):
            new_value = getattr(r_obj, self.link_attr) or []
            new_value.remove(old_link_value)
            new_value.append(new_link_value)
        else:
            new_value = new_link_value
        # Update remote object link attribute and save it
        setattr(r_obj, self.link_attr, new_value)
        if not r_obj.save():
            log.error("%s : An error occured updating attr %s of the object %s with value %s to update link value of object %s from '%s' to '%s'", self, self.link_attr, r_obj, new_value, obj, old_link_value, new_link_value)
            return False
        log.debug('%s : object %s renamed on it relation with %s', self, obj, r_obj)
        return True

    def handle_object_changes(self, old_obj, new_obj):
        """
        Handle changes on object

        :param  old_obj:    The object in its state before modification
        :param  new_obj:    The object in its state after modification
        """

        # Retreive old link_value
        try:
            old_link_value = self._get_link_value(old_obj)
        except LDAPRelationLinkAttributeValueMissing:
            log.warning('No link value retreived from old object')
            return True

        # Retreive new link_value
        try:
            new_link_value = self._get_link_value(new_obj)
        except LDAPRelationLinkAttributeValueMissing:
            log.info('No link value retreived from new object : remove relation with related objects')

        # If no change, stop
        if old_link_value == new_link_value:
            log.debug('%s : No change detected on link value', self)
            return True

        log.debug('%s : change detected on link value : "%s" => "%s"', self, old_link_value, new_link_value)

        # Link modified : handle link_value change
        related_objs = self.get_related_objects(old_obj)
        if not related_objs:
            log.debug('%s : object %s is currently not link with other object', self, old_obj)
            return True

        if self.multiple:
            # On multiple relation case
            error = False
            for r_obj_id, r_obj in related_objs.items():    # pylint: disable=unused-variable
                if new_link_value is None:
                    if not self.remove_related_object(new_obj, r_obj, link_value=old_link_value):
                        error = True
                elif not self.rename_obj_on_related_object(new_obj, r_obj, old_link_value, new_link_value):
                    error = True
            return not error

        # On non multiple relation case
        if new_link_value is None:
            return self.remove_related_object(new_obj, related_objs, link_value=old_link_value)
        return self.rename_obj_on_related_object(new_obj, related_objs, old_link_value, new_link_value)

    def handle_object_removal(self, obj):
        """
        Handle removal of object

        :param  obj:    The removed object
        """
        # Retreive old link_value
        try:
            link_value = self._get_link_value(obj)
        except LDAPRelationLinkAttributeValueMissing:
            log.warning('No link value retreived from object')
            return True

        related_objs = self.get_related_objects(obj)
        if not related_objs:
            log.debug('%s : object %s is not link with other object', self, obj)
            return True
        log.debug('%s : object %s is link with some other object(s) : %s', self, obj, related_objs)

        if self.multiple:
            # On multiple relation case
            error = False
            for r_obj_id, r_obj in related_objs.items():    # pylint: disable=unused-variable
                if not self.remove_related_object(obj, r_obj, link_value=link_value):
                    error = True

            return not error

        # On non multiple relation case
        return self.remove_related_object(obj, related_objs, link_value=link_value)
